
#include <config.h>

#define DEFINE_GLOBALS

#include "messages.h"
#include "ftpd_p.h"
#include "dynamic.h"
#include "ftpwho-update.h"
#include "globals.h"
#include "caps.h"
#if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
# include "upload-pipe.h"
#endif
#ifdef WITH_DMALLOC
# include <dmalloc.h>
#endif

#ifndef HAVE_SYS_FSUID_H
static void disablesignals(void)
{
	sigset_t sigs;
	
	sigfillset(&sigs);
	if (sigprocmask(SIG_BLOCK, &sigs, &old_sigmask) < 0) {
		_EXIT(EXIT_FAILURE);
	}
}

static void enablesignals(void)
{
	if (sigprocmask(SIG_SETMASK, &old_sigmask, NULL) < 0) {
		_EXIT(EXIT_FAILURE);
	}
}

void usleep2(const unsigned long microsec)
{
	disablesignals();
	usleep(microsec);
	enablesignals();
}
#endif

void die(const int err, const int facility, const char * const format, ...)
{
	va_list va;
	char line[MAX_SYSLOG_LINE];

	va_start(va, format);
#ifndef HAVE_SYS_FSUID_H
	disablesignals();
#endif		
	vsnprintf(line, sizeof line, format, va);
	printf("%d %s", err, line);
	puts("\r");
	fflush(stdout);
	logfile(facility, line);
	va_end(va);
	_EXIT(-facility-1);
}

void die_mem(void)
{
	die(421, LOG_ERR, MSG_OUT_OF_MEMORY);
}

RETSIGTYPE sigurg(int sig)
{
	const int olderrno = errno;
	
	(void) sig;
	if (xferfd < 0) {
		return;
	}
	close(xferfd);
	xferfd = -1;
#ifndef HAVE_SYS_FSUID_H
	disablesignals();
#endif	
	addreply(426, MSG_ABORTED);
	doreply();
	sfgets();
	addreply(226, MSG_ABORTED);
	doreply();
#ifndef HAVE_SYS_FSUID_H
	enablesignals();
#endif		
	errno = olderrno;
}

RETSIGTYPE sigalarm(int sig)
{
	(void) sig;
	die(421, LOG_INFO, MSG_TIMEOUT);
}

#ifndef NO_STANDALONE
RETSIGTYPE sigchild(int sig)
{
	const int olderrno = errno;
	pid_t pid;

	(void) sig;
#ifdef HAVE_WAITPID
	while ((pid = waitpid((pid_t) -1, NULL, WNOHANG)) > (pid_t) 0) {
		if (nb_children > 0U) {
			nb_children--;
		}
		iptrack_delete_pid(pid);
	}
#else
	while ((pid = wait3(NULL, WNOHANG, NULL)) > (pid_t) 0) {
		if (nb_children > 0U) {		
			nb_children--;
		}
		iptrack_delete_pid(pid);
	}
#endif
	errno = olderrno;
}
#endif

RETSIGTYPE sigterm_client(int sig)
{
	(void) sig;
	
	_EXIT(EXIT_SUCCESS);
}

#ifndef NO_STANDALONE
RETSIGTYPE sigterm(int sig)
{
	const int olderrno = errno;
	(void) sig;
	
	stop_server = 1;
	if (listenfd >= 0) {
		shutdown(listenfd, 2);
		close(listenfd);		
	}
	errno = olderrno;
}

static void set_cloexec_flag(const int fd)
{
	fcntl(fd, F_SETFD, FD_CLOEXEC);
}
#endif

static void clearargs(char **argv) {
#if defined(__linux__) && !defined(HAVE_SETPROCTITLE)
	if (*argv != NULL) {
		argv++;
	}
	while (*argv != NULL) {		
		memset(*argv, 0, strlen(*argv));
		argv++;
	}
#else
	(void) argv;
#endif
}

void setprogname(const char * const title)
{
#ifdef HAVE_SETPROCTITLE
	setproctitle("-%s", title);
#elif defined(__linux__)
	if (prg_name != NULL) {
		strncpy(prg_name, title, 31);
		prg_name[31] = 0;		
	}
#else
	(void) title;
#endif
}

/* Check it an address is valid, return 1 if ok, 0 otherwise.
 * Unfortunately, multicasting with the FTP protocol is impossible,
 * you have to use things like MTP instead. So prohibit multicast.
 */

int checkvalidaddr(const struct sockaddr_storage * const addr)
{
	if (addr == NULL) {
		return 0;
	}
	if (STORAGE_FAMILY(*addr) == AF_INET6) {
		if (IN6_IS_ADDR_MULTICAST(&STORAGE_SIN_ADDR6_NF(*addr)) ||
			IN6_IS_ADDR_UNSPECIFIED(&STORAGE_SIN_ADDR6_NF(*addr))) {
			return 0;
		}
		return 1;
	} else if (STORAGE_FAMILY(*addr) == AF_INET) {
		if (ntohl(STORAGE_SIN_ADDR(*addr)) == INADDR_ANY ||
			ntohl(STORAGE_SIN_ADDR(*addr)) == INADDR_NONE ||
			ntohl(STORAGE_SIN_ADDR(*addr)) == INADDR_BROADCAST ||
			IN_MULTICAST(ntohl(STORAGE_SIN_ADDR(*addr))) ||
			IN_BADCLASS(ntohl(STORAGE_SIN_ADDR(*addr)))) {
			return 0;
		}
		return 1;
	}
	return 0;
}

/* Convert a 4-in-6 address in pure IPv4 */

void fourinsix(struct sockaddr_storage *v6)
{
	struct sockaddr_storage v4;
		
	if (v6ready == 0 || STORAGE_FAMILY(*v6) != AF_INET6 ||
		IN6_IS_ADDR_V4MAPPED(&STORAGE_SIN_ADDR6_NF(*v6)) == 0) {
		return;
	}
	memset(&v4, 0, sizeof v4);
	STORAGE_FAMILY(v4) = AF_INET;
	memcpy(&STORAGE_SIN_ADDR(v4),
		   (unsigned char *) &STORAGE_SIN_ADDR6(*v6) + 12,
		   sizeof STORAGE_SIN_ADDR(v4));
	STORAGE_PORT(v4) = STORAGE_PORT6(*v6);
	SET_STORAGE_LEN(v4, sizeof(struct sockaddr_in));
	*v6 = v4;
}		

/* Return 0 if s1 == s2 , 1 if s1 != s2 , -1 if error */

int addrcmp(const struct sockaddr_storage * const s1,
			const struct sockaddr_storage * const s2)
{	
	if (STORAGE_FAMILY(*s1) == AF_INET6) {
		if (STORAGE_FAMILY(*s2) != AF_INET6) {
			return 1;
		}
		if (IN6_ARE_ADDR_EQUAL(&STORAGE_SIN_ADDR6_NF(*s1), &STORAGE_SIN_ADDR6_NF(*s2))) {
			return 0;
		} else {
			return 1;		
		}
	} else if (STORAGE_FAMILY(*s1) == AF_INET) {
		if (STORAGE_FAMILY(*s2) != AF_INET) {
			return 1;
		}
		if (STORAGE_SIN_ADDR(*s1) == STORAGE_SIN_ADDR(*s2)) {
			return 0;
		} else {
			return 1;
		}
	}
	return -1;
}

int generic_aton(const char *src, struct sockaddr_storage *a)
{
	if (inet_pton(AF_INET6, src, &STORAGE_SIN_ADDR6(*a)) > 0) {
		STORAGE_FAMILY(*a) = AF_INET6;
		return 0;
	}
	if (inet_pton(AF_INET, src, &STORAGE_SIN_ADDR(*a)) > 0) {
		STORAGE_FAMILY(*a) = AF_INET;		
		return 0;
	}
	memset(a, 0, sizeof *a);
	
	return -1;
}

void logfile(const int crit, const char *format, ...)
{
#ifdef NON_ROOT_FTP
	(void) crit;
	(void) format;
#else
	va_list va;
	const char *urgency;
	char line[MAX_SYSLOG_LINE];	
	
	if (no_syslog != 0) {
		return;
	}
	va_start(va, format);
	vsnprintf(line, sizeof line, format, va);
	line[sizeof line - 1U] = 0;	
	switch (crit) {
	case LOG_INFO:
		urgency = "[INFO] ";
		break;
	case LOG_WARNING:
		urgency = "[WARNING] ";
		break;
	case LOG_ERR:
		urgency = "[ERROR] ";
		break;
	default:
		urgency = "";
	}
# ifdef SAVE_DESCRIPTORS
	openlog("pure-ftpd", 0, syslog_facility);
# endif
	syslog(crit, "(%s@%s) %s%s",
		   ((loggedin != 0 && *account != 0) ? account : "?"),
		   (*host != 0 ? host : "?"),
		   urgency, line);
# ifdef SAVE_DESCRIPTORS
	closelog();
# endif
	va_end(va);
#endif
}

struct passwd *getpwnam_unix(const char * const name)
{
	if (chrooted != 0) {
		return NULL;
	}
	return getpwnam(name);
}

#ifndef NO_STANDALONE

/* this is taken from the code examples for Stevens' "Advanced
 * Programming in the Unix Environment. The code is publicly available
 * at ftp://ftp.uu.net/published/books/stevens.advprog.tar.Z */

static int open_max(void)
{
	int openmax;
	
	if ((openmax = sysconf(_SC_OPEN_MAX)) < 0) {
		perror("_SC_OPEN_MAX");
		_EXIT(EXIT_FAILURE);
	}
	return openmax;
}

#endif

void addreply_noformat(int code, const char *line)
{
	struct reply *p;

	if (code) {
		replycode = code;
	}
	p =	(struct reply *) malloc(offsetof(struct reply, line) +
								strlen(line) + 1U);
	if (!p) {
		die_mem();
	}
	strcpy(p->line, line);	/* Okay, no overflow */
	p->prev = lastreply;
	lastreply = p;
}

void addreply(int code, const char *line, ...)
{
	struct reply *p;
	struct reply *q;
	va_list ap;
	int offs;
	char buf[PATH_MAX + 50U];	

	if (code) {
		replycode = code;
	}
	q = NULL;

	va_start(ap, line);
	if (vsnprintf(buf, sizeof buf, line, ap) < 0) {
		_EXIT(EXIT_FAILURE);
	}
	va_end(ap);

	offs = strlen(buf);
	while (offs && isspace(buf[offs]))
		buf[offs--] = 0;
	if (!offs && buf[0])
		offs++;
	while (offs) {
		while (offs && buf[offs] != '\n' && isspace(buf[offs]))
			buf[offs--] = 0;
		while (offs && buf[offs] != '\n')
			offs--;
		if (offs)
			offs++;
		/* Fixed off-by-one -J. */
		p =	(struct reply *) malloc(offsetof(struct reply, line) +
									strlen(buf + offs) + 1U);
		if (!p) {
			die_mem();
		}
		strcpy(p->line, buf + offs);	/* Okay, no overflow */
		if (q) {
			p->prev = q->prev;
			q->prev = p;
		} else {
			p->prev = lastreply;
			lastreply = p;
		}
		q = p;
		if (offs)
			buf[--offs] = '\r';
	}
}

void replycont(struct reply *p)
{
	if (p) {
		replycont(p->prev);
		CORK_ON(1);
		printf("%3d-%s\r\n", replycode, p->line);
		if (logging > 1)
			logfile(LOG_DEBUG, "%3d-%s\n", replycode, p->line);
		free(p);
	}
}

void doreply(void)
{
	if (lastreply) {
		replycont(lastreply->prev);
		printf("%3d %s\r\n", replycode, lastreply->line);
		CORK_OFF(1);
		if (logging > 1)
			logfile(LOG_DEBUG, "%3d %s\n", replycode, lastreply->line);
		free((char *) lastreply);
		lastreply = NULL;
	}
	fflush(stdout);
}

/* Check if a file name is valid. Files names starting
 * with a dot are only allowed to root and to users
 * chroot()ed in their home directories -Jedi. */

static int checknamesanity(const char *name, int dot_ok)
{
	register const char *namepnt;

	if (name == NULL || *name == 0) {
		return -1;
	}
    /* optimize . and .. */
    if (name[0] == '.' && (name[1] == 0 || (name[1] == '.' && name[2] == 0))) {
        return 0;
    }
    namepnt = name;
    while (*namepnt != 0) {
        if ((unsigned char) *namepnt < 32U || *namepnt == '\\') {
            return -1;
        }
        if (dot_ok == 0) {
            if (*namepnt == '/') {
                namepnt++;
            } else if (namepnt != name) {
                namepnt++;
                continue;
            }
            if (namepnt[0] == 0) {     /* /$ */
                return 0;
            }
            if (namepnt[0] == '.') {   /* /. */
                if (namepnt[1] == 0) { /* /.$ => ok */
                    return 0;
                }
                if (namepnt[1] == '.') {   /* /.. */
                    if (namepnt[2] == 0) {   /* /..$ => ok */
                        return 0;
                    }
                    if (namepnt[2] != '/') {   /* /..[^/] => *NO* */
                        return -1;
                    }
                } else if (namepnt[1] != '/') {   /* /.[^/]/ => *NO* */
                    return -1;
                }
            }
            if (namepnt != name) {
                continue;
            }
        }
        namepnt++;        
    }
    return 0;
}

void do_ipv6_port(char *p, char delim)
{
	struct sockaddr_storage a;
	char *deb;

	deb = p;	
	while (*p && strchr("0123456789abcdefABCDEF:", *p) != NULL) {
		p++;
	}
	if (*p != delim || atoi(p + 1) == 0) {
		nope:
		close(datafd);
		datafd = -1;		
		addreply(501, MSG_SYNTAX_ERROR_IP);
		return;
	}		
	*p++ = 0;
	if (generic_aton(deb, &a) != 0) {
		goto nope;
	}
	doport2(a, atoi(p));
}

void doeprt(char *p)
{
	char delim;
	int family;

	delim = *p++;
	family = atoi(p);
	while (isdigit(*p)) {
		p++;
    }
	if (*p == delim)
		p++;
	else {
		addreply(501, MSG_SYNTAX_ERROR_IP);
		return;
	}
	if (family == 2 && v6ready) {
		do_ipv6_port(p, delim);
		return;
	}
	if (family != 1) {
		if (v6ready) {
			addreply(522, MSG_ONLY_IPV4V6);
		} else {
			addreply(522, MSG_ONLY_IPV4);
        }
		return;
	}

	{
		unsigned int a1, a2, a3, a4, port = 0;
		/* there should be dot-decimal ip as rfc2428 states,
		 * but troll used for some reason "comma-decimal" notation
		 * so i decided to leave it */
		if ((sscanf(p, "%u,%u,%u,%u", &a1, &a2, &a3, &a4) != 4 &&
			 sscanf(p, "%u.%u.%u.%u", &a1, &a2, &a3, &a4) != 4) ||
			a1 > 255 || a2 > 255 || a3 > 255 || a4 > 255 || 
			(a1|a2|a3|a4) == 0) {
			addreply(501, MSG_SYNTAX_ERROR_IP);
			return;
		}
		while (*p && strchr("0123456789.,", *p)) {
			p++;
        }
		if (*p == delim) {
			port = atoi(++p);
			while (*p && isdigit(*p)) {
				p++;
            }
		}
		if (*p != delim || port > 65535U || port <= 0U) {
			addreply(501, MSG_SYNTAX_ERROR_IP);
			return;
		} else {
			struct sockaddr_storage a;
			
			memset(&a, 0, sizeof a);
			STORAGE_FAMILY(a) = AF_INET;
			STORAGE_SIN_ADDR(a) =
				htonl((a1 << 24) | (a2 << 16) | (a3 << 8) | a4);
			SET_STORAGE_LEN(a, sizeof(struct sockaddr_in));
			doport2(a, port);
		}
	}
}

#ifndef MINIMAL

/* small help routine to display a banner */
void dobanner(void)
{
	FILE *msg;
	if ((msg = fopen(".banner", "rt")) != NULL
# ifdef WITH_WELCOME_MSG
	|| (msg = fopen("welcome.msg", "rt")) != NULL
# endif
		) {
		/* if you think this is too small, send me mail.
		 NOTE: It should not be unlimited. */
		char buffer[2000];
        char *bufpnt = buffer;
		size_t len = fread(buffer, (size_t) 1U, sizeof buffer - 1U, msg);
		fclose(msg);
		if (len > (size_t) 0U && len < (sizeof buffer - 1U)) {
			buffer[len] = 0;
            while (*bufpnt == '\n') {
                bufpnt++;
            }
			addreply(0, "%s", bufpnt);
		}
	}
}

#endif

#ifndef MINIMAL

int modernformat(const char *file,
				 char *target, size_t target_size)
{
	struct stat st;
	const char *ft;
	struct tm *t;
	int ret = 0;

	if (stat(file, &st) != 0 || 
		!(t = gmtime((time_t *) &st.st_mtime))) {
		return -1;
	}
	if (S_ISREG(st.st_mode)) {
		ft = "file";
	} else if (S_ISDIR(st.st_mode)) {
		ret = 1;
		ft = "dir";
		if (*file == '.') {
			if (file[1] == '.' && file[2] == 0) {
				ft = "pdir";
			} else if (file[1] == 0) {
				ft = "cdir";
			} 
		} else if (*file == '/' && file[1] == 0) {
			ft = "pdir";
		}
	} else {
		ft = "unknown";
	}
	if (guest != 0) {
		if (snprintf(target, target_size,
					 "type=%s;siz%c=%llu;modify=%04d%02d%02d%02d%02d%02d;UNIX.mode=0%o;unique=%xg%llx %s",
					 ft, 
					 ret ? 'd' : 'e',
					 (unsigned long long) st.st_size,
					 t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
					 t->tm_hour, t->tm_min, t->tm_sec,
					 (unsigned int) st.st_mode & 07777,
					 (unsigned int) st.st_dev,
					 (unsigned long long) st.st_ino,   
					 file) < 0) {
			_EXIT(EXIT_FAILURE);
		}
	} else {
		if (snprintf(target, target_size,
					 "type=%s;siz%c=%llu;modify=%04d%02d%02d%02d%02d%02d;UNIX.mode=0%o;UNIX.uid=%lld;UNIX.gid=%lld;unique=%xg%llx %s",
					 ft, 
					 ret ? 'd' : 'e',
					 (unsigned long long) st.st_size,
					 t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
					 t->tm_hour, t->tm_min, t->tm_sec,
					 (unsigned int) st.st_mode & 07777,
					 (unsigned long long) st.st_uid,
					 (unsigned long long) st.st_gid,					 
					 (unsigned int) st.st_dev,
					 (unsigned long long) st.st_ino,   
					 file) < 0) {
			_EXIT(EXIT_FAILURE);
		}		
	}
	return ret;
}

#define MLST_BEGIN "Begin" CRLF

void domlst(const char * const file)
{
	char line[PATH_MAX + 256U] = MLST_BEGIN;
	
	if (modernformat(file, line + (sizeof MLST_BEGIN - 1U),
					 sizeof line - (sizeof MLST_BEGIN - 1U)) >= 0) {
		addreply_noformat(0, line);
		addreply(250, "End.");
	} else {
		addreply(550, MSG_STAT_FAILURE2);
	}
}

void donoop(void)
{
	addreply(200, MSG_SLEEPING);
}

#endif

void douser(const char *username)
{
	struct passwd *pw;

	if (loggedin) {
		if (username) {
			if (!guest) {
				addreply(530, MSG_ALREADY_LOGGED);
			} else if (broken) {
				addreply(331, MSG_ANY_PASSWORD);
			} else {
				addreply(230, MSG_ANONYMOUS_LOGGED);
				dot_ok = 0;
				dot_read_forbidden = !dot_read_anon_ok;
				dot_write_forbidden = 1;
			}
		}
		return;
	}
	if (anon_only <= 0 && username != NULL &&
		strcasecmp(username, "ftp") && strcasecmp(username, "anonymous")) {
		strncpy(account, username, sizeof(account) - 1);
		account[sizeof(account) - 1] = 0;
		addreply(331, MSG_USER_OK, account);
		loggedin = 0;
	} else if (anon_only < 0) {		   /* anonymous, but -E */
        if (broken != 0) {
            addreply(331, MSG_USER_OK, username);
            return;
        } else {
            die(530, LOG_DEBUG, MSG_NO_ANONYMOUS_LOGIN);
        }
	} else {
#ifdef WITH_VIRTUAL_HOSTS
		char name[PATH_MAX + 1U];
		char hbuf[NI_MAXHOST];
#endif
		if (chrooted != 0) {
			die(421, LOG_DEBUG, MSG_CANT_DO_TWICE);
		}
#ifdef NON_ROOT_FTP
		rd = "/";
#else
		if ((pw = getpwnam_wrapper("ftp")) == NULL ||
			pw->pw_uid == 0 || pw->pw_gid == 0 ||
			setgroups(sizeof(pw->pw_gid) / sizeof(gid_t), &(pw->pw_gid)) ||
			setgid(pw->pw_gid) || setegid(pw->pw_gid)) {
			cantsec:
			die(421, LOG_ERR, MSG_UNABLE_SECURE_ANON);            
        }
# ifdef WITH_VIRTUAL_HOSTS
        getnameinfo((struct sockaddr *) &ctrlconn, STORAGE_LEN(ctrlconn),
                    hbuf, sizeof hbuf, NULL, (size_t) 0U, NI_NUMERICHOST);
		if (snprintf(name, sizeof name, VHOST_PATH "/%s/.", hbuf) <= 0) {
			_EXIT(EXIT_FAILURE);
		}        
        if (chdir(name) != 0) {        /* non-virtual */
# endif
            if ((rd = strdup(pw->pw_dir)) == NULL ||
                chdir(rd) || chroot(rd) || chdir("/")) {
                goto cantsec;
            }
            rd_len = strlen(rd) + 1U;
            logfile(LOG_INFO, MSG_ANONYMOUS_LOGGED);
# ifdef WITH_VIRTUAL_HOSTS
        } else {                       /* virtual host */
			rd_len = strlen(hbuf) + 3U;/* 3, if we ever add a / (paranoia) */
			if ((rd = malloc(rd_len)) == NULL ||
                chdir(name) || chroot(name) || chdir("/") ||
                snprintf(rd, rd_len, "%s:", hbuf) <= 0) {
                goto cantsec;
            }
			logfile(LOG_INFO, MSG_ANONYMOUS_LOGGED_VIRTUAL ": %s", hbuf);            
        }
# endif
#endif
		chrooted = 1;
#ifndef NON_ROOT_FTP		
# ifndef HAVE_SYS_FSUID_H
		user_uid = pw->pw_uid;
		user_gid = pw->pw_gid;		
# endif
		if (pw->pw_uid > (uid_t) 0) {
# ifdef HAVE_SYS_FSUID_H
			setfsgid(pw->pw_gid);			
			setfsuid(pw->pw_uid);
# else
			if (seteuid(user_uid)) {
				goto cantsec;
			}
# endif
		}
#endif
#ifdef USE_CAPABILITIES
		drop_login_caps();
#endif
#ifdef THROTTLING
		if (throttling != 0) {
			addreply(0, MSG_BANDWIDTH_RESTRICTED);
			nice(NICE_VALUE);
		} else {
			throttling_delay = throttling_bandwidth_ul =
				throttling_bandwidth_dl = 0UL;
		}
#endif
#ifndef MINIMAL
		dobanner();
#endif
		if (broken) {
			addreply(331, MSG_ANONYMOUS_ANY_PASSWORD);
		} else {
			addreply(230, MSG_ANONYMOUS_LOGGED);
		}
		dot_ok = 0;
		dot_write_forbidden = 1;
		dot_read_forbidden = !dot_read_anon_ok;
		strncpy(account, "ftp", sizeof account - 1U);
		account[(sizeof account) - 1U] = 0;
#ifdef FTPWHO
		if (shm_data_cur != NULL) {
			ftpwho_lock();
			strncpy(shm_data_cur->account, account, 
					sizeof shm_data_cur->account - 1U);
			shm_data_cur->account[sizeof shm_data_cur->account - 1U] = 0;
			ftpwho_unlock();
			state_needs_update = 1;
		}
#endif
		loggedin = guest = 1;
	}
	if (getcwd(wd, sizeof wd) == NULL) {
		wd[0] = '/';
		wd[1] = 0;
	}
}

#ifndef USE_PAM
static int pw_unix_check(const char *account, const char *password)
{
	const char *cpwd = NULL;
	struct passwd *pw;
#ifdef USE_SHADOW
	struct spwd *spw;
#endif

	if ((pw = getpwnam_wrapper(account)) == NULL) {
		return 0;
	}
#ifdef USE_SHADOW
	if (!strcmp(pw->pw_passwd, "x") && (spw = getspnam(account))) {
		cpwd = spw->sp_pwdp[0] == '@' ? "*" : spw->sp_pwdp;
		if (spw->sp_expire > 0 || spw->sp_max > 0) {
			long today = time(NULL) / (24L * 60L * 60L);

			if (spw->sp_expire > 0 && spw->sp_expire < today) {
				return 0;			   /* account expired */
			}
			if (spw->sp_max > 0 && spw->sp_lstchg > 0 &&
				(spw->sp_lstchg + spw->sp_max < today)) {
				return 0;			   /* password expired */
			}
		}
	} else
#endif
	{
		cpwd = pw->pw_passwd;
	}

	return cpwd && !strcmp(cpwd, (const char *) crypt(password, cpwd));
}
#endif

static int pw_check(const char *account, const char *password)
{
#ifdef NON_ROOT_FTP
	return 1;
#else
# ifdef USE_PAM
	return pam_ftp_check(account, password);
# else
	return pw_unix_check(account, password);
# endif
#endif
}

void dopass(const char *password)
{
	struct passwd *pw;
	int ngroups;
	char *shell;
	gid_t g[NGROUPS_MAX];	
	static unsigned int tapping;

	if (loggedin != 0) {
		if (guest != 0) {
			addreply(230, MSG_NO_PASSWORD_NEEDED);
		} else {
			addreply(530, MSG_NOTRUST);
		}
		return;
	}
	if (*account == 0) {
		addreply(530, MSG_WHOAREYOU);
		return;
	}
	if (pw_check(account, password) == 0) {
		addreply(530, MSG_AUTH_FAILED);
		doreply();
		if (tapping >= MAX_PASSWD_TRIES) {
			logfile(LOG_ERR, MSG_AUTH_TOOMANY);
			_EXIT(EXIT_FAILURE);
		}
		logfile(LOG_WARNING, MSG_AUTH_FAILED_LOG, account);
		usleep2(tapping * PASSWD_FAILURE_DELAY);
		tapping++;
		return;
	}

	pw = getpwnam_wrapper(account);

#ifdef USE_PAM
	shell = pw ? pw->pw_shell : NULL;
#else
	setusershell();
	while ((shell = (char *) getusershell()) != NULL &&
		   pw && strcmp(pw->pw_shell, shell));
	endusershell();
#endif

	if (shell && pw && pw->pw_uid >= useruid) {
		/* handle /home/user/./public_html form */
		/* XXX /etc/ftpusers */
		char *hd;
		
		if ((rd = strdup(pw->pw_dir)) == NULL) {
			die_mem();
		}
		rd_len = strlen(rd) + 1U;
		hd = strstr(rd, "/./");
		if (hd) {
			*hd = 0;
			hd += 2;
			if (chrooted != 0) {
				die(421, LOG_DEBUG, MSG_CANT_DO_TWICE);
			}
			if (chdir(rd) || chroot(rd) || chdir(hd)) {
				die(421, LOG_ERR, MSG_NO_HOMEDIR);
			}
			chrooted = 1;
#ifdef RATIOS
			if (ratio_for_non_anon == 0) {
				ratio_upload = ratio_download = 0U;
			}
#endif
		} else {
			(void) free(rd);
			rd = (char *) "/";
			rd_len = 2U;
			if (chdir(pw->pw_dir)) {
				/* non-chrooted users can see everything, so let them
				   in anyway */
				if (userchroot == 0) {
					addreply(0, MSG_NO_HOMEDIR2 " [%s].",
							 pw->pw_dir, strerror(errno));
					addreply(0, MSG_START_SLASH);
					chdir("/");
				} else {
					die(421, LOG_ERR, MSG_NO_HOMEDIR);
				}
			}
		}
		if (getcwd(wd, sizeof wd) == NULL) {
			wd[0] = '/';
			wd[1] = 0;
		}
#ifdef HAVE_SYS_FSUID_H
		setfsuid(pw->pw_uid);
#endif
#ifndef NON_ROOT_FTP
		setgroups(sizeof(pw->pw_gid) / sizeof(gid_t), &(pw->pw_gid));
		if (setgid(pw->pw_gid) ||
			setegid(pw->pw_gid)) {
			_EXIT(EXIT_FAILURE);
		}
# ifdef HAVE_INITGROUPS
		initgroups(pw->pw_name, pw->pw_gid);
# endif		
# ifndef HAVE_SYS_FSUID_H
		user_uid = pw->pw_uid;
		user_gid = pw->pw_gid;	
		if (seteuid(user_uid)) {
			_EXIT(EXIT_FAILURE);
		}
# endif	
#endif		
		/* never chroot and ratio our beloved sysadmin */
		if (pw->pw_uid == (uid_t) 0) {			
			dot_write_forbidden = dot_read_forbidden = 0;
#ifdef RATIOS
			ratio_upload = ratio_download = 0U;
#endif
			userchroot = 0;		
		} else if (userchroot == 1 && chrooted == 0) {
			/* check if user belongs to the trusted group id */
			if (pw->pw_gid == chroot_trustedgid) {
				userchroot = 0;
				dot_write_forbidden = dot_read_forbidden = 0;				
#ifdef RATIOS
				ratio_upload = ratio_download = 0U;
#endif
			} else {
				gid_t *suppgroups;
				int n, n2;

				if ((n = getgroups(0, NULL)) > 0) {
					if ((suppgroups =
						 ALLOCA(n * (sizeof *suppgroups))) == NULL) {
						die_mem();
					}
					n2 = getgroups(n, suppgroups);
					/* Jedi's paranoia */
					if (n2 < n) {
						n = n2;
					}
					while (n != 0) {
						n--;
						if (suppgroups[n] == chroot_trustedgid) {
							userchroot = 0;
							dot_write_forbidden = dot_read_forbidden = 0;
#ifdef RATIOS
							ratio_upload = ratio_download = 0U;
#endif
							break;
						}
					};
					ALLOCA_FREE(suppgroups);
				}
			}
		}
	} else {
		addreply(530, MSG_NOTRUST);
		return;
	}
	if (loggedin == 0) {
		candownload = 1;		/* real users can always download */
	}
#ifdef THROTTLING
	if ((throttling == 2) || (guest != 0 && throttling == 1)) {
		addreply(0, MSG_BANDWIDTH_RESTRICTED);
		nice(NICE_VALUE);
	} else {
		throttling_delay = throttling_bandwidth_dl = 
			throttling_bandwidth_ul = 0UL;
	}
#endif
	ngroups = getgroups(NGROUPS_MAX, g);
	if (!guest && ngroups > 0) {
		size_t p;
		const char *q;
		char reply[80];		

		if (snprintf(reply, sizeof reply, MSG_USER_GROUP_ACCESS ": ", account) <= 0) {
			_EXIT(EXIT_FAILURE);
		}
		p = strlen(reply);
		do {
			ngroups--;
			if ((q = getgroup(g[ngroups])) == NULL) {
				continue;
			}
			if (p + strlen(q) > 75) {
				reply[p] = 0;
				addreply(0, "%s", reply);
				*reply = 0;
				p = (size_t) 0U;
			}
			reply[p++] = ' ';
			while (*q != 0 && p < sizeof reply) {
				reply[p++] = *q++;
			}
		} while (ngroups > 0);
		reply[p] = 0;
		addreply(0, "%s", reply);
	}
	if (guest == 0 && allowfxp == 1) {
		addreply(0, MSG_FXP_SUPPORT);
	}
#ifdef RATIOS
	if (ratio_upload > 0) {
		addreply(0, MSG_RATIO, ratio_upload, ratio_download);
	}
#endif
	if (userchroot != 0 && chrooted == 0) {
#ifndef NON_ROOT_FTP
# ifndef HAVE_SYS_FSUID_H
	disablesignals();
	seteuid((uid_t) 0);		
# endif
#endif
		if (chdir(wd)
#ifndef NON_ROOT_FTP
			|| chroot(wd)
#endif
			) {	/* should never fail */
			die(421, LOG_ERR, MSG_CHROOT_FAILED);
		}
#ifndef NON_ROOT_FTP
# ifndef HAVE_SYS_FSUID_H
		if (seteuid(user_uid) != 0) {
			_EXIT(EXIT_FAILURE);
		}
		enablesignals();
# endif
#endif		
		chrooted = 1;
#ifdef RATIOS		
		if (ratio_for_non_anon == 0) {
			ratio_upload = ratio_download = 0U;
		}	
#endif
		if ((rd = strdup(wd)) == NULL) {
			die_mem();
		}
		wd[0] = '/';
		wd[1] = 0;
		if (chdir(wd)) {
			_EXIT(EXIT_FAILURE);			
		}
		addreply(230, MSG_CURRENT_RESTRICTED_DIR_IS, wd);
	} else {
		addreply(230, MSG_CURRENT_DIR_IS, wd);
	}
		
	dot_ok = 1;	
#ifdef USE_CAPABILITIES
	drop_login_caps();
#endif
	logfile(LOG_INFO, MSG_IS_NOW_LOGGED_IN, account);
# ifdef FTPWHO
	if (shm_data_cur != NULL) {
		ftpwho_lock();
		strncpy(shm_data_cur->account, account, 
				sizeof shm_data_cur->account - 1U);
		shm_data_cur->account[sizeof shm_data_cur->account - 1U] = 0;
		ftpwho_unlock();
		state_needs_update = 1;
	}
# endif				
	loggedin = 1;
	if (getcwd(wd, sizeof wd) == NULL) {
		wd[0] = '/';
		wd[1] = 0;
	}
#ifndef MINIMAL
	dobanner();
#endif	
}

#ifdef WITH_LDAP
# define IGNORE_TILDE
#endif	

void docwd(const char *dir)
{
	const char *where;
	char buffer[PATH_MAX + 256U];

#ifndef IGNORE_TILDE
	if (loggedin && !guest && dir && *dir == '~') {
		if (chrooted != 0) {
			buffer[0] = '/';
			strncpy(buffer + 1, dir + 1, sizeof buffer - 2U);
			buffer[sizeof buffer - 1U] = 0;
		} else {
			struct passwd *pw;
			size_t i;
			const char *p;

			i = 0U;
			p = dir;
			p++;
			/* If sizeof(cmd) < sizeof(buffer) we can't get any
			 * buffer overflow here. But be paranoid anyway, and
			 * check for i < sizeof buffer. -J. */
			while (*p && *p != '/' && i < sizeof buffer - 1U) {
				buffer[i++] = *p++;
			}
			buffer[i] = 0;

			if ((pw = getpwnam_wrapper(i ? buffer : account))) {
				if (snprintf(buffer, sizeof buffer - 1U, "%s%s", pw->pw_dir, p) <= 0) {
					_EXIT(EXIT_FAILURE);				
				}
			} else {
				*buffer = 0;
			}
		}
	} else {
#endif
		*buffer = 0;
#ifndef IGNORE_TILDE
	}
#endif
	where = *buffer ? buffer : (dir && *dir ? dir : "/");
    if (checknamesanity(where, dot_read_forbidden != 0 ? 0 : dot_ok) != 0) {
        addreply(550, MSG_SANITY_FILE_FAILURE, where);
        return;
    }
	if (chdir(where)) {
		if (snprintf(buffer, sizeof buffer,
					 MSG_CANT_CHANGE_DIR ": %s",
					 dir, strerror(errno)) <= 0) {
			_EXIT(EXIT_FAILURE);							
		}
		logfile(LOG_INFO, "%s", buffer);
		addreply(530, "%s", buffer);
		return;
	}
	
#ifndef MINIMAL
	{
		FILE *msg;

		if (sizeof buffer > (size_t) 2000U && 
			(msg = fopen(".message", "rt")) != NULL) {
			size_t len = fread(buffer, (size_t) 1U, (size_t) 2000U, msg);
			fclose(msg);
			if (len > (size_t) 0U && len < (size_t) 2000U) {
				buffer[len] = 0;
				addreply(0, "%s", buffer);
			}
		}
	}
#endif
	if (getcwd(wd, sizeof wd) == NULL) {
		if (*dir == '/') {
			if (snprintf(wd, sizeof wd, "%s", dir) <= 0) { /* already checked */
				_EXIT(EXIT_FAILURE);
			}
		} else {
			if (snprintf(wd, sizeof wd, "%s/%s", wd, dir) <= 0) {
				die(421, LOG_ERR, MSG_PATH_TOO_LONG);
			}
		}
	}	
	addreply(250, MSG_CURRENT_DIR_IS, wd);	
}

static void iptropize(const struct sockaddr_storage *ss)
{
	size_t t = sizeof *ss;
	register const unsigned char *s = (const unsigned char *) ss;
	
	iptropy = getpid();
	do {
		iptropy += (iptropy << 5);
		iptropy ^= (int) *s++;
	} while (--t > 0U);
}

static int zrand(void)
{
	int fd;
	int ret;
	
	if ((fd = open("/dev/urandom", O_RDONLY)) < 0 &&
		(fd = open("/dev/random", O_RDONLY)) < 0) {
		nax:
		return rand() ^ iptropy;
	}
	if (read(fd, &ret, sizeof ret) != sizeof ret) {
		close(fd);
		goto nax;
	}
	close(fd);	
	return ret ^ iptropy;
}

static void nokeepalive(const int fd)
{
#ifdef SO_KEEPALIVE
    {
		int fodder = 0;
    
		setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *) &fodder, sizeof fodder);
    }
#endif
#ifdef SO_LINGER
        {
            struct linger li;
            li.l_onoff = li.l_linger = 0;
            setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *) &li, sizeof li);
        }
#endif
}

/* psvtype = 0 : PASV */
/* psvtype = 1 : EPSV */
/* psvtype = 2 : SPSV */

void dopasv(int psvtype)
{
	socklen_t socksize;
	unsigned long a = 0U;
	unsigned int p;
	unsigned int on;
	struct sockaddr_storage dataconn;	/* my data connection endpoint */
	unsigned int firstporttried;
	struct sockaddr_storage ctrlconn2;

	if (loggedin == 0) {
		addreply(501, MSG_NOT_LOGGED_IN);
		return;
	}
	if (datafd >= 0) {				/* for buggy clients */
		close(datafd);
		datafd = -1;
	}
	fourinsix(&ctrlconn);
	if (STORAGE_FAMILY(ctrlconn) == AF_INET6 && psvtype == 0) {
		addreply(425, MSG_CANT_PASV);
		return;
	}
    firstporttried = firstport + zrand() % (lastport - firstport + 1);
	p = firstporttried;
	datafd = socket(STORAGE_FAMILY(ctrlconn), SOCK_STREAM, IPPROTO_TCP);
	if (datafd < 0) {
		error(425, MSG_CANT_PASSIVE);
		return;
	}
	on = 1;
	if (setsockopt(datafd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on) < 0) {
		error(421, "setsockopt");
		return;
	}
	ctrlconn2 = ctrlconn;
	for (;;) {
		if (STORAGE_FAMILY(ctrlconn2) == AF_INET6) {
			STORAGE_PORT6(ctrlconn2) = htons(p);
		} else {
			STORAGE_PORT(ctrlconn2) = htons(p);
		}
		if (bind(datafd, (struct sockaddr *) &ctrlconn2, 
				 STORAGE_LEN(ctrlconn2)) == 0) {
			break;
		}
        p--;
        if (p < firstport) {
            p = lastport;
        }
		if (p == firstporttried) {
			close(datafd);
			datafd = -1;			
            addreply(425, MSG_PORTS_BUSY);
			return;
		}
	}
	alarm(idletime);
	socksize = (socklen_t) sizeof dataconn;
	if (listen(datafd, DEFAULT_BACKLOG_DATA) < 0 ||
	    getsockname(datafd, (struct sockaddr *) &dataconn, &socksize) < 0) {
		close(datafd);
		datafd = -1;		
		error(425, MSG_GETSOCKNAME_DATA);
		return;
	}
	fourinsix(&dataconn);
	if (checkvalidaddr(&dataconn) == 0) {
		close(datafd);
		datafd = -1;		
		addreply(425, MSG_INVALID_IP);
		return;					
	}
	if (STORAGE_FAMILY(dataconn) == AF_INET6) {
		p = ntohs((unsigned short int) STORAGE_PORT6(dataconn));
	} else {
		p = ntohs((unsigned short int) STORAGE_PORT(dataconn));
    }
	switch (psvtype) {
	case 0 :
		if (STORAGE_FAMILY(force_passive_ip) == 0) {			
			a = ntohl(STORAGE_SIN_ADDR(dataconn));
		} else if (STORAGE_FAMILY(force_passive_ip) == AF_INET6) {
			close(datafd);
			datafd = -1;			
			addreply(425, MSG_NO_EPSV);
			return;			
		} else if (STORAGE_FAMILY(force_passive_ip) == AF_INET) {
			a = ntohl(STORAGE_SIN_ADDR(force_passive_ip));
		} else {
			_EXIT(EXIT_FAILURE);			
		}
		
/* According to RFC, any message can follow 227. But broken NAT gateways
 * and connection tracking code rely on this. So don't translate the following
 * messages */

		addreply(227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d).",
				 (a >> 24) & 255, (a >> 16) & 255, (a >> 8) & 255, a & 255,
				 (p >> 8) & 255, p & 255);
		break;
	case 1 :
		addreply(229, "Extended Passive mode OK (|||%d|)", p);
		break;
	case 2 :
		addreply(227, "%d", p);
		break;
	default :
		_EXIT(EXIT_FAILURE);					
	}
	passive = 1;
	
	return;
}

void doport(const char *arg)
{
	unsigned int a1, a2, a3, a4, p1, p2;
	struct sockaddr_storage a;
	
	if (6 != sscanf(arg, "%u,%u,%u,%u,%u,%u",
					&a1, &a2, &a3, &a4, &p1, &p2) ||
		a1 > 255 || a2 > 255 || a3 > 255 || a4 > 255 ||
		p1 > 255 || p2 > 255 || (a1|a2|a3|a4) == 0 ||
		(p1|p2) == 0) {
		addreply(501, MSG_SYNTAX_ERROR_IP);
		return;
	}
	memset(&a, 0, sizeof a);
	STORAGE_FAMILY(a) = AF_INET;
	STORAGE_SIN_ADDR(a) =
		htonl((a1 << 24) | (a2 << 16) | (a3 << 8) | a4);
	SET_STORAGE_LEN(a, sizeof(struct sockaddr_in));
	doport2(a, (p1 << 8) | p2);
}

void doport2(struct sockaddr_storage a, unsigned int p)
{
	struct sockaddr_storage dataconn;  /* his endpoint */
	int on;

	if (loggedin == 0) {
		addreply(501, MSG_NOT_LOGGED_IN);
		return;
	}
	if (epsv_all != 0) {
		addreply(501, MSG_ACTIVE_DISABLED);
		return;		
	}
	if (datafd >= 0) {				/* for buggy clients saying PORT over and over */
		close(datafd);
		datafd = -1;
	}
	if (p < 1024) {
		addreply(501, MSG_BAD_PORT);
		return;
	}
#ifndef NON_ROOT_FTP
# ifndef HAVE_SYS_FSUID_H
	disablesignals();
	seteuid((uid_t) 0);
# endif
#endif	
	if (STORAGE_FAMILY(a) == AF_INET6) {
		datafd = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
	} else {
		datafd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	}
	if (datafd < 0) {
		data_socket_error:
#ifndef NON_ROOT_FTP
# ifndef HAVE_SYS_FSUID_H
		if (seteuid(user_uid) != 0) {
			_EXIT(EXIT_FAILURE);
		}
		enablesignals();
# endif
#endif
		close(datafd);
		datafd = -1;				
		error(425, MSG_CANT_CREATE_DATA_SOCKET);
		return;
	}
	on = 1;
	setsockopt(datafd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	memcpy(&dataconn, &ctrlconn, sizeof dataconn);
#ifndef NON_ROOT_FTP
	if (STORAGE_FAMILY(dataconn) == AF_INET6) {
		STORAGE_PORT6(dataconn) = htons((short) DEFAULT_FTP_DATA_PORT);
	} else {
		STORAGE_PORT(dataconn) = htons((short) DEFAULT_FTP_DATA_PORT);
	}
#else
	if (STORAGE_FAMILY(dataconn) == AF_INET6) {
		STORAGE_PORT6(dataconn) = htons((short) (0));
	} else {
		STORAGE_PORT(dataconn) = htons((short) (0));
	}
#endif
	if (bind(datafd, (struct sockaddr *) &dataconn, STORAGE_LEN(dataconn)) < 0) {
		goto data_socket_error;
	}
#ifndef NON_ROOT_FTP
# ifndef HAVE_SYS_FSUID_H
	if (seteuid(user_uid) != 0) {
		_EXIT(EXIT_FAILURE);
	}
	enablesignals();
# endif	
#endif
	peerdataport = p;
	if (addrcmp(&a, &peer) != 0) {
		char hbuf[NI_MAXHOST];
		char peerbuf[NI_MAXHOST];			
		
		getnameinfo((struct sockaddr *) &a, STORAGE_LEN(a),
					hbuf, sizeof hbuf, NULL, (size_t) 0U, NI_NUMERICHOST);
		getnameinfo((struct sockaddr *) &peer, STORAGE_LEN(peer),
					peerbuf, sizeof peerbuf, NULL, (size_t) 0U, NI_NUMERICHOST);		
		if (allowfxp == 0 || (allowfxp == 1 && guest != 0)) {
			close(datafd);
			datafd = -1;			
			addreply(500, MSG_NO_FXP, hbuf, peerbuf);
			return;
		} else {
			addreply(0, MSG_FXP, peerbuf, hbuf);
			memcpy(&peer, &a, sizeof a);
		}
	}
	passive = 0;

	addreply(200, MSG_PORT_SUCCESSFUL);
	return;
}

int opendata(void)
{
	struct sockaddr_storage dataconn;	/* his data connection endpoint */
	int fd;
	socklen_t socksize;

	if (datafd < 0) {
		error(425, MSG_NO_DATA_CONN);
		return -1;
	}
	
	if (passive) {
		fd_set rs;
		struct timeval tv;

		alarm(idletime);
		
		for (;;) {
			FD_ZERO(&rs);
			FD_SET(datafd, &rs);
			tv.tv_sec = idletime;
			tv.tv_usec = 0;
			/* I suppose it would be better to listen for ABRT too... */
			
			if (select(datafd + 1, &rs, NULL, NULL, &tv) <= 0) {
				die(421, LOG_INFO, MSG_TIMEOUT_DATA , (unsigned long) idletime);
			}
			socksize = (socklen_t) sizeof(dataconn);
			memset(&dataconn, 0, sizeof dataconn);
			fd = accept(datafd, (struct sockaddr *) &dataconn, &socksize);
			if (fd < 0) {
				nope:
				close(datafd);
				datafd = -1;				
				error(421, MSG_ACCEPT_FAILED);
				return -1;
			}
			if (STORAGE_FAMILY(dataconn) != AF_INET
			    && STORAGE_FAMILY(dataconn) != AF_INET6) {
				close(fd);
				goto nope;
			}
			fourinsix(&dataconn);
			if (addrcmp(&peer, &dataconn) == 0) {
				break;
			}
			if (allowfxp == 0 || (allowfxp == 1 && guest != 0)) {
				shutdown(fd, 2);
				close(fd);
			} else {
				break;
			}
		} 		
		addreply(150, MSG_ACCEPT_SUCCESS);
	} else {
		struct sockaddr_storage peer2;
		unsigned long tries = 1UL + idletime / 2UL;

		peer2 = peer;
		if (STORAGE_FAMILY(peer) == AF_INET6) {
			STORAGE_PORT6(peer2) = htons(peerdataport);
		} else {
			STORAGE_PORT(peer2) = htons(peerdataport);
		}
		again:
		if (connect(datafd, (struct sockaddr *) &peer2, STORAGE_LEN(peer2))) {
			if (errno == EAGAIN && tries > 0UL) {
				tries--;
				usleep2(1000000UL);
				goto again;
			}
			addreply(425, MSG_CNX_PORT_FAILED ": %s",
					 peerdataport, strerror(errno));            
			close(datafd);
			datafd = -1;			
			return -1;
		}
		fd = datafd;
		datafd = -1;
		addreply(150, MSG_CNX_PORT, peerdataport);
	}

	{
		int fodder;
#ifdef IPTOS_THROUGHPUT
		fodder = IPTOS_THROUGHPUT;
		setsockopt(fd, SOL_IP, IP_TOS, (char *) &fodder, sizeof fodder);
#endif
#ifndef ENABLE_NAGLE
# ifdef TCP_NODELAY
	   fodder = 1;
	   setsockopt(fd, SOL_TCP, TCP_NODELAY, (char *) &fodder, sizeof fodder);
# endif
#endif
#ifdef SO_SNDBUF
		fodder = window;
		setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &fodder, sizeof fodder);
#endif
#ifdef SO_RCVBUF
		fodder = window;			/* not that important, but... */
		setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &fodder, sizeof fodder);
#endif
        nokeepalive(fd);
	}
	
	return fd;
}

#ifndef MINIMAL
void dochmod(char *name, const mode_t mode)
{
	if (guest != 0) {
		addreply(550, MSG_ANON_CANT_CHANGE_PERMS);
		return;
	}
	if (name == NULL || *name == 0) {
		addreply(501, MSG_NO_FILE_NAME);
		return;
	}
    if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
        addreply(550, MSG_SANITY_FILE_FAILURE);
        return;
    }    
    if (chmod(name, mode) < 0) {
        addreply(550, MSG_CHMOD_FAILED ": %s", name, strerror(errno));
        return;
    } 
    addreply(200, MSG_CHMOD_SUCCESS, name);
}
#endif

void dodele(char *name)
{
	if (guest != 0) {
		addreply(550, MSG_ANON_CANT_DELETE);
		return;
	}
	if (name == NULL || *name == 0) {
		addreply(501, MSG_NO_FILE_NAME);
		return;
	}		
    if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
        addreply(550, MSG_SANITY_FILE_FAILURE);
        return;
    }
    if (unlink(name) < 0) {
        addreply(550, MSG_DELE_FAILED ": %s", name, strerror(errno));
        return;
    }     
    addreply(250, MSG_DELE_SUCCESS, name);
}

#ifndef WITHOUT_ASCII	
static ssize_t dowrite(int fd, const unsigned char *buff, size_t size)
{
	/* XXX This function would be faster if we did our own scan... */

	unsigned char *v = memchr(buff, '\n', size);
	unsigned char *w = memchr(buff, '\r', size);
	if ((w && !v) || (w && (w < v))) {
		/* Found \r */
		ssize_t r = write(fd, buff, w - buff);
		if (r != (w - buff))
			return r;
		switch (write(fd, "\r\0", 2)) {	/* OHOH, 1 byte lost... ;-) */
		case 1:				/* CRAP! */
			return w - buff;	/* XXX Unrecoverable ERROR without further state info! */
		case 2:				/* OK */
			return w - buff + 1;
		default:				/* Return the byte before, so \r\n will be sent again */
			return w - buff;
		}
	} else if (v) {
		/* \n found */
		ssize_t r = write(fd, buff, v - buff);
		if (r != (v - buff))
			return r;
		switch (write(fd, "\r\n", 2)) {	/* OHOH, 1 bytes lost... ;-) */
		case 1:				/* CRAP! */
			return v - buff;	/* XXX Unrecoverable ERROR without further state info! */
		case 2:				/* OK */
			return v - buff + 1;
		default:				/* Return the byte before, so \r\n will be sent again */
			return v - buff;
		}
	} else {
		return write(fd, buff, size);
	}
}
#endif

static void displayrate(const char *word, off_t size, 
						const struct timeval * const started,
						const char * const name, int up)
{
	struct timeval ended;
	double t;
	double speed;
	char speedstring[64];	

	(void) gettimeofday(&ended, NULL);
	
	t = (ended.tv_sec + ended.tv_usec / 1000000.0) -
		(started->tv_sec + started->tv_usec / 1000000.0);
	addreply(226, MSG_TRANSFER_SUCCESSFUL);
	if (t > 0.0 && size > 2 * window) {
		speed = (size - window) / t;
		if (speed < 0.0) {
			speed = size / t;
		}
	} else {
		speed = 0.0;
	}
	if (speed > 524288.0) {
		addreply(0, MSG_TRANSFER_RATE_M, t, speed / 1048576.0);
	} else if (speed > 512.0) {
		addreply(0, MSG_TRANSFER_RATE_K, t, speed / 1024.0);
 	} else if (speed > 0.1) {
		addreply(0, MSG_TRANSFER_RATE_B, t, speed);
    }
	if (snprintf(speedstring, sizeof speedstring, " (%llu bytes, %.2fKB/sec)",
				 (unsigned long long) size, speed / 1024.0) > 0) {
        logfile(LOG_INFO, "%s%s%s%s %s %s", rd ? rd : "",
                *name == '/' ? "" : wd,
                (*name != '/' && (!*wd || wd[strlen(wd) - 1] != '/'))
                ? "/" : "", name, word, speedstring);
    }
#if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
	if (do_upload_script != 0 && up != 0) {
		char filename_for_pipe[MAXPATHLEN + 32U];
		char path[MAXPATHLEN + 1U];
		
		if (realpath(name, path) == NULL) {
			return;
		}
		if (snprintf(filename_for_pipe, sizeof filename_for_pipe,
					 "\001%s%s%s", (rd ? rd : ""),
					 (rd ? ((rd[strlen(rd) - 1] == '/') ? "" : "/") : "/"),
					 (*path == '/' ? path + 1 : path)) < 0) {
			return;
		}					 					 
		upload_pipe_push(filename_for_pipe);
	}
#else
	(void) up;
#endif
}

static void displayopenfailure(const char * const name)
{
    char buffer[PATH_MAX + 42U];
    
    if (snprintf(buffer, sizeof buffer, MSG_OPEN_FAILURE, name) <= 0) {
        _EXIT(EXIT_FAILURE);
    }
    error(550, buffer);   
}

void doretr(char *name)
{
	int c, f, s, skip;
	off_t o;
	struct stat st;
	char *p, *buf;
	int left;
	struct timeval started;
#ifdef THROTTLING
	struct timeval ended;
	off_t transmitted = 0;
#endif

	if (!candownload) {
		addreply(550, MSG_LOAD_TOO_HIGH, load);
        goto end;
	}
    if (type < 1 || (type == 1 && restartat > (off_t) 1)) {
        addreply(503, MSG_NO_ASCII_RESUME);
        goto end;
    }
	if (checknamesanity(name, dot_read_forbidden != 0 ? 0 : dot_ok) != 0) {
		addreply(550, MSG_SANITY_FILE_FAILURE);
        goto end;
	}
	f = open(name, O_RDONLY);
	if (f < 0) {
        displayopenfailure(name);
        goto end;
	}
	if (fstat(f, &st) < 0) {
        stat_failure:
		close(f);
		error(451, MSG_STAT_FAILURE);
        goto end;
	}
	if (S_ISLNK(st.st_mode)) {
        if (stat(name, &st) < 0) {
            goto stat_failure;
        }
    }    
	if (restartat && (restartat > st.st_size)) {
        close(f);
		addreply(451, MSG_REST_TOO_LARGE_FOR_FILE "\n" MSG_REST_RESET, 
				 (long long) restartat, (long long) st.st_size);
        goto end;        
	}
    if (!S_ISREG(st.st_mode)
#ifndef WITH_LARGE_FILES
		|| st.st_size >= INT_MAX
#endif
		) {
		close(f);
		addreply(450, MSG_NOT_REGULAR_FILE);
        goto end;
	}
	if (warez && (st.st_uid == warez) && guest) {
		close(f);
		addreply(550, MSG_NOT_MODERATED);
        goto end;
	}
#ifdef RATIOS
	if (ratio_upload > 0U) {
		if ((downloaded + st.st_size - restartat) / ratio_download >
			(uploaded / ratio_upload)) {
			close(f);
			addreply(550, MSG_RATIO_DENIAL, ratio_upload, ratio_download, 
					 (unsigned long long) uploaded / 1024ULL,
					 (unsigned long long) downloaded / 1024ULL);
            goto end;
		}
	}
#endif
	c = opendata();
	if (c < 0) {
		close(f);
        goto end;
	}
	xferfd = c;
	if (restartat > st.st_size) {
		/* some clients insist on doing this.  I can't imagine why. */
		addreply(226, MSG_NO_MORE_TO_DOWNLOAD);
		close(f);
		xferfd = -1;
		close(c);
        goto end;
	}
#ifdef NON_BLOCKING_DATA_SOCKET
	if ((s = fcntl(c, F_GETFL, 0)) < 0) {
		close(f);
		xferfd = -1;
		close(c);
		error(451, "fcntl");		
		goto end;
	}
	s |= FNDELAY;
	fcntl(c, F_SETFL, s);
#endif

#ifndef DISABLE_HUMOR
    if ((time(NULL) % 100) == 0) {
        addreply(0, MSG_WINNER);
    }
#endif
	if (st.st_size - restartat > 4096) {
		addreply(0, MSG_KBYTES_LEFT, (double) ((st.st_size - restartat) / 1024.0));
	}
	doreply();
	state_needs_update = 1;
	setprogname("pure-ftpd [DOWNLOAD]");
			
#ifdef FTPWHO
	if (shm_data_cur != NULL) {
		const size_t sl = strlen(name);
		
		ftpwho_lock();
		shm_data_cur->state = FTPWHO_STATE_DOWNLOAD;
		shm_data_cur->download_total_size = st.st_size;
		shm_data_cur->download_current_size = restartat;
		shm_data_cur->restartat = restartat;
		(void) time(&shm_data_cur->xfer_date);
		if (sl < sizeof shm_data_cur->filename - 1U) {
			/* no overflow, see the previous line */
			strcpy(shm_data_cur->filename, name);   /* audited - ok */
		} else {
			/* same thing here, no possible buffer overflow, keep cool */
			strcpy(shm_data_cur->filename,   /* audited - ok */
				   &name[sl - sizeof shm_data_cur->filename + 1U]);   
		}	
		ftpwho_unlock();
	}
#endif	

	(void) gettimeofday(&started, NULL);
	
#ifndef WITHOUT_ASCII
	if (type == 2)
#endif
	{			/* Binary */
		o = restartat;
		CORK_ON(c);
#ifndef SENDFILE_NONE
		while (o < st.st_size) {
# ifdef FTPWHO
			/* There is no locking here, and it is intentional */
            if (shm_data_cur != NULL) {
                shm_data_cur->download_current_size = o;
            }
# endif
			left = st.st_size - o;
# ifdef FTPWHO
			if (left > MAP_SIZE) {
				left = MAP_SIZE;
			}			
# elif defined(THROTTLING)			
			if (throttling_bandwidth_dl > 0UL && left > MAP_SIZE) {
				left = MAP_SIZE;				
			}
# endif			
			while (left > 0) {
# ifdef SENDFILE_LINUX				
				ssize_t w;
# elif defined(SENDFILE_FREEBSD)
				int w;
# else
#  error Configuration error
# endif
				alarm(idletime);
# ifdef SENDFILE_LINUX
				w = sendfile(c, f, &o, (size_t) left);
# elif defined(SENDFILE_FREEBSD)
				{
					off_t written;
					
					if ((w = sendfile(f, c, o, (size_t) left, NULL, &written, 0)) < 0) {
						if (errno == EAGAIN && written > 0) {
							w = 0;
							o += written;
						}
					} else {
						o += written;
					}					
				}
# endif
				if (w < (ssize_t) 0) {
					if (errno == EAGAIN) {
						/* wait idletime seconds for progress */
						fd_set rs;
						fd_set ws;
						struct timeval tv;
						
						FD_ZERO(&rs);
						FD_ZERO(&ws);
						FD_SET(0, &rs);
						FD_SET(c, &ws);
						tv.tv_sec = idletime;
						tv.tv_usec = 0;
						select(c + 1, &rs, &ws, NULL, &tv);
						if (FD_ISSET(0, &rs)) {
							/* we assume is is ABRT since nothing else is legal */
							close(f);
							xferfd = -1;
							close(c);
							addreply(426, MSG_ABORTED);
							goto end;
						} else if (!(FD_ISSET(c, &ws))) {
							/* client presumably gone away */
							xferfd = -1;
							close(c);
							die(421, LOG_INFO, MSG_TIMEOUT_DATA ,
								(unsigned long) idletime);
						}
						w = 0;
					} else {
						close(f);						
						if (xferfd >= 0) {
							xferfd = -1;
							close(c);										
							error(450, MSG_DATA_WRITE_FAILED);
						}
						goto end;
					}
				}
				downloaded += w;
				if (!w) {		/* Maybe the file has shrunk? */
					if (fstat(f, &st) < 0) {
						o = st.st_size;
					}
					left = 0;
				} else {
					left -= w;
# ifdef THROTTLING
					if (o < st.st_size && throttling_bandwidth_dl > 0UL) {
						long double delay;
						(void) gettimeofday(&ended, NULL);
						transmitted += w;
						delay = (transmitted / (long double) throttling_bandwidth_dl)
							- (long double) (ended.tv_sec - started.tv_sec);
						if (delay > (long double) MAX_THROTTLING_DELAY) {
							(void) gettimeofday(&started, NULL);
							transmitted = (off_t) 0;
                            delay = (long double) MAX_THROTTLING_DELAY;
						}
                        if (delay > 0.0L) {
							usleep2(delay * 1000000.0L);
						}
					}
# endif														
				}
			}
		}
#else
		o = restartat & ~(MAP_SIZE - 1);
		skip = restartat - o;
		while (o < st.st_size) {
# ifdef FTPWHO
            if (shm_data_cur != NULL) {
                shm_data_cur->download_current_size = o;
            }
# endif
			
			left = st.st_size - o;
			if (left > MAP_SIZE) {
				left = MAP_SIZE;
			}
			buf = mmap(0, left, PROT_READ, MAP_FILE | MAP_SHARED, f, o);
			if (buf == MAP_FAILED) {
				close(c);				
				close(f);
				xferfd = -1;
				error(451, MSG_MMAP_FAILED);				
                goto end;
			}
			p = buf;
			o += left;
			s = left;
			while (left > skip) {
				ssize_t w;
				
				w = write(c, p + skip, (size_t) (left - skip));
				if (w < (ssize_t) 0) {
					if (errno == EAGAIN) {
						/* wait idletime seconds for progress */
						fd_set rs;
						fd_set ws;
						struct timeval tv;
						
						FD_ZERO(&rs);
						FD_ZERO(&ws);
						FD_SET(0, &rs);
						FD_SET(c, &ws);
						tv.tv_sec = idletime;
						tv.tv_usec = 0;
						select(c + 1, &rs, &ws, NULL, &tv);
						if (FD_ISSET( 0, &rs )) {
							/* we assume is is ABRT since nothing else is legal */
							munmap(buf, s);
							close(f);
							xferfd = -1;
							close(c);
							addreply(426, MSG_ABORTED);
							goto end;
						} else if (!(FD_ISSET(c, &ws))) {
							/* client presumably gone away */
							die(421, LOG_INFO, MSG_TIMEOUT_DATA ,
								(unsigned long) idletime);
						}
						w = (ssize_t) 0;
					} else {
						close(f);						
						if (xferfd >= 0) {
							xferfd = -1;
							close(c);									
							error(450, MSG_DATA_WRITE_FAILED);
						}
						goto end;
					}
				}
				downloaded += w;
				left -= w;
				p += w;                
# ifdef THROTTLING
				if (left > skip && throttling_bandwidth_dl > 0UL) {
					long double delay;
					
					(void) gettimeofday(&ended, NULL);
					transmitted += w;
					delay = (transmitted / (long double) throttling_bandwidth_dl)
						- (long double) (ended.tv_sec - started.tv_sec);
					if (delay > (long double) MAX_THROTTLING_DELAY) {
						(void) gettimeofday(&started, NULL);
						transmitted = (off_t) 0;
                        delay = (long double) MAX_THROTTLING_DELAY;
					}
                    if (delay > 0.0L) {
						usleep2(delay * 1000000.0L);
					}					
				}
# endif									
			}
			skip = 0;
		    munmap(buf, s);
		}
#endif
	}
#ifndef WITHOUT_ASCII	
	else {					/* ASCII */
		o = restartat & ~(MAP_SIZE - 1);
		skip = restartat - o;
		CORK_ON(c);
		while (o < st.st_size) {
# ifdef FTPWHO
            if (shm_data_cur != NULL) {
                shm_data_cur->download_current_size = o;
            }
# endif			
			left = st.st_size - o;
			if (left > MAP_SIZE)
				left = MAP_SIZE;
			buf = mmap(0, left, PROT_READ, MAP_FILE | MAP_SHARED, f, o);
			if (buf == MAP_FAILED) {
				close(c);				
				close(f);
				xferfd = -1;
				error(451, MSG_MMAP_FAILED);				
                goto end;
			}
			p = buf;
			o += left;
			s = left;
			while (left > skip) {
				ssize_t w;
				alarm(idletime);
				w = dowrite(c, (const unsigned char *) p + skip, (size_t) (left - skip));
				
				if (w < (ssize_t) 0) {
					if (errno == EAGAIN) {
						/* wait idletime seconds for progress */
						fd_set rs;
						fd_set ws;
						struct timeval tv;

						FD_ZERO(&rs);
						FD_ZERO(&ws);
						FD_SET(0, &rs);
						FD_SET(c, &ws);
						tv.tv_sec = idletime;
						tv.tv_usec = 0;
						select(c + 1, &rs, &ws, NULL, &tv);
						if (FD_ISSET(0, &rs)) {
							/* we assume is is ABRT since nothing else is legal */
							munmap(buf, s);
							close(c);							
							close(f);
							xferfd = -1;
							addreply(426, MSG_ABORTED);
							goto end;
						} else if (!(FD_ISSET(c, &ws))) {
							/* client presumably gone away */
							die(421, LOG_INFO, MSG_TIMEOUT_DATA ,
								   (unsigned long) idletime);
						}
						w = (ssize_t) 0;
					} else {
						close(f);						
						if (xferfd >= 0) {
							xferfd = -1;
							close(c);	
							error(450, MSG_DATA_WRITE_FAILED);
						}
						goto end;
					}
				}
				downloaded += w;
				left -= w;
				p += w;                
# ifdef THROTTLING
				if (left > skip && throttling_bandwidth_dl > 0UL) {
					long double delay;
					
					(void) gettimeofday(&ended, NULL);
					transmitted += w;
					delay = (transmitted / (long double) throttling_bandwidth_dl)
						- (long double) (ended.tv_sec - started.tv_sec);
					if (delay > (long double) MAX_THROTTLING_DELAY) {
						(void) gettimeofday(&started, NULL);
						transmitted = (off_t) 0;
                        delay = (long double) MAX_THROTTLING_DELAY;                        
					}
                    if (delay > 0.0L) {
						usleep2(delay * 1000000.0L);
					}
				}
# endif
			}
			skip = 0;
			munmap(buf, s);
		}
	}
#endif
	close(f);
	xferfd = -1;
	close(c);
    displayrate(MSG_DOWNLOADED, st.st_size - restartat, &started, name, 0);
	end:
	restartat = (off_t) 0;
}

void dorest(const char *name)
{
	char *endptr;
	restartat = strtoul(name, &endptr, 10);
	if (*endptr) {
		restartat = 0;
		addreply(501, MSG_REST_NOT_NUMERIC "\n" MSG_REST_RESET);
	} else {
		if (type == 1 && restartat != 0) {
#ifdef STRICT_REST
			addreply(504, MSG_REST_ASCII_STRICT);
#else
			addreply(350, MSG_REST_ASCII_WORKAROUND,
					 (long long) restartat);			
#endif
		} else {
			if (restartat != 0) {
				logfile(LOG_NOTICE, MSG_REST_SUCCESS, (long long) restartat);
			}
			addreply(350, MSG_REST_SUCCESS, (long long) restartat);
		}
	}
}

void domkd(char *name)
{
	if (guest != 0 && allow_anon_mkdir == 0) {
		addreply(550, MSG_ANON_CANT_MKD);
		return;
	}
	if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
		addreply(550, MSG_SANITY_DIRECTORY_FAILURE);
		return;
	}
	if ((mkdir(name, 0777 & ~u_mask_d)) < 0) {
		error(550, MSG_MKD_FAILURE);
	} else {
		addreply(257, "\"%s\" : " MSG_MKD_SUCCESS, name);
	}
}

void dormd(char *name)
{
	if (guest) {
		addreply(550, MSG_ANON_CANT_RMD);
		return;
	}
	if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
		addreply(550, MSG_SANITY_DIRECTORY_FAILURE);
		return;
	}	
	if ((rmdir(name)) < 0) {
		error(550, MSG_RMD_FAILURE);
	} else {
		addreply(250, MSG_RMD_SUCCESS);
	}
}

#ifndef MINIMAL
void dofeat(void)
{
# define FEAT  "Extensions supported:" CRLF \
    " EPRT" CRLF " IDLE" CRLF " MDTM" CRLF " SIZE" CRLF \
    " REST STREAM" CRLF " MLST" CRLF " MLSD"
# ifdef DEBUG
#  define FEAT_DEBUG CRLF " XDBG"
# else
#  define FEAT_DEBUG ""
# endif
# define FEAT_PASV CRLF " PASV" CRLF " EPSV" CRLF " SPSV"
    char feat[] = FEAT FEAT_DEBUG FEAT_PASV;
        
    if (disallow_passive != 0) {
        feat[sizeof FEAT FEAT_DEBUG] = 0;
    }
    addreply_noformat(0,  feat);    
	addreply(211, "End.");
}
#endif

#ifndef MINIMAL
void dostou(void)
{
	static unsigned int seq = 0U;
	struct timeval tv;
	struct timezone tz;
	char file[PATH_MAX + 1U];
	
	if (gettimeofday(&tv, &tz) != 0) {
		error(553, MSG_TIMESTAMP_FAILURE);
		return;	
	}
	if (snprintf(file, sizeof file, "pureftpd.%08lx.%02lx.%04x",
				 (unsigned long) tv.tv_sec,
				 (unsigned long) tv.tv_usec & 0xff,
				 seq) <= 0) {
		_EXIT(EXIT_FAILURE);
	}
	seq++;
	seq &= 0xffff;
	addreply(150, "FILE: %s", file);
	dostor(file, 0, 1);
}
#endif

static int tryautorename(char * const name)
{
    char name2[MAXPATHLEN];
    unsigned int gc = 0U;
    int f;
    
    for (;;) {
        gc++;
        if (gc == 0U || 
            snprintf(name2, sizeof name2, "%s.%u", name, gc) < 0) {
            error(553, MSG_OPEN_FAILURE2);
            return -1;
        }
        /* O_EXCL avoids races - great */
        if ((f = open(name2, O_CREAT | O_EXCL | O_WRONLY, 0777 & ~u_mask)) >= 0) {
            strncpy(name, name2, MAXPATHLEN);
            name[MAXPATHLEN - 1U] = 0;   /* useless, but paranoid */
            return f;
        }
    }
}

void dostor(char *name, const int append, const int autorename)
{
	int c, f;
	char *p;
	char buf[window];			/* Thanks to GCC */
	int r;
	off_t filesize = (off_t) 0U;
	STATFS_STRUCT statfsbuf;
	struct stat st;
	struct timeval started;
#ifdef THROTTLING
	struct timeval ended;
	off_t transmitted = 0;
#endif
    int unlinkret = -1;

	if (type < 1 || (type == 1 && restartat > (off_t) 1)) {
		addreply(503, MSG_NO_ASCII_RESUME);
        goto end;
	}
    if (guest != 0 && anon_noupload != 0) {
        addreply(553, MSG_ANON_CANT_OVERWRITE);
        goto end;
    }
	if (maxdiskusagepct > 0.0 && STATFS(name, &statfsbuf) == 0 &&
        (double) STATFS_BLOCKS(statfsbuf) > 0.0) {
		double jam;

		jam = (double) STATFS_BAVAIL(statfsbuf) / 
			(double) STATFS_BLOCKS(statfsbuf);
		if (jam < maxdiskusagepct) {
			addreply(553, MSG_NO_DISK_SPACE);
            goto end;
		}
	}    
	if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
		addreply(553, MSG_SANITY_FILE_FAILURE);
        goto end;
	}
    if (autorename != 0 && restartat == (off_t) 0) {
        if ((f = tryautorename(name)) < 0) {
            goto end;
        }
    } else {                           /* no auto rename */
		if ((f = open(name, O_CREAT | O_WRONLY, 0777 & ~u_mask)) < 0) {
            error(553, MSG_OPEN_FAILURE2);
            goto end;
        }
        if (fstat(f, &st) < 0) {
            close(f);
            error(553, MSG_STAT_FAILURE2);
            goto end;
        }
        /* Anonymous users *CAN* overwrite 0-bytes files - This is the right behavior */
        if (st.st_size > (off_t) 0) {
            if (guest != 0) {
                addreply(553, MSG_ANON_CANT_OVERWRITE);
                close(f);
                goto end;
            }
            if (append != 0) {
                restartat = st.st_size;
            } 
        } else {
            restartat = (off_t) 0;
        }
        if (restartat > st.st_size) {
            restartat = st.st_size;
        }
        if (restartat > (off_t) 0 && lseek(f, restartat, SEEK_SET) < 0) {
            close(f);                
            error(451, "seek");
            goto end;
        }
        if (restartat < st.st_size) {
            if (ftruncate(f, restartat) < 0) {
                close(f);                    
                error(451, "ftruncate");
                goto end;
            }
        }
    }    
	c = opendata();
	if (c < 0) {
		close(f);
        goto end;
	}
	xferfd = c;
	doreply();
	state_needs_update = 1;
	setprogname("pure-ftpd [UPLOAD]");
	filesize = restartat;
	
#ifdef FTPWHO
	if (shm_data_cur != NULL) {
		const size_t sl = strlen(name);
		
		ftpwho_lock();
		shm_data_cur->state = FTPWHO_STATE_UPLOAD;
		shm_data_cur->download_total_size = (off_t) 0U;
		shm_data_cur->download_current_size = (off_t) filesize;
		shm_data_cur->restartat = restartat;		
		(void) time(&shm_data_cur->xfer_date);		
		if (sl < sizeof shm_data_cur->filename - 1U) {
			/* no overflow, see the previous line */
			strcpy(shm_data_cur->filename, name);   /* audited - ok */
		} else {
			/* same thing here, no possible buffer overflow, keep cool */
			strcpy(shm_data_cur->filename,   /* audited - ok */
				   &name[sl - sizeof shm_data_cur->filename + 1U]);
		}	
		ftpwho_unlock();
	}
#endif
	
	(void) gettimeofday(&started, NULL);
	do {
		/* wait idletime seconds for data to be available */
		fd_set rs;
		struct timeval tv;

		FD_ZERO(&rs);
		FD_SET(0, &rs);
		FD_SET(c, &rs);
		tv.tv_sec = idletime;
		tv.tv_usec = 0;
		select(c + 1, &rs, NULL, NULL, &tv);
		if (FD_ISSET(0, &rs)) {
			close(f);
			xferfd = -1;
			close(c);
			addreply(0, MSG_ABRT_ONLY);
			addreply(426, MSG_ABORTED);
            if (guest != 0) {
                unlinkret = unlink(name);
            }
			addreply(0, "%s %s", name,
					 unlinkret ? MSG_UPLOAD_PARTIAL : MSG_REMOVED);
            goto end;
		} else if (!(FD_ISSET(c, &rs))) {
			/* client presumably gone away */
            if (guest != 0) {
                unlink(name);
            }
			die(421, LOG_INFO, MSG_TIMEOUT_DATA , (unsigned long) idletime);
		}
		alarm(idletime);
		r = read(c, buf, (size_t) window);      /* window = sizeof buf */
		if (r > (ssize_t) 0) {
			p = buf;

#ifdef THROTTLING
			if (throttling_bandwidth_ul > 0UL) {
				long double delay;
				
				(void) gettimeofday(&ended, NULL);
				transmitted += r;
				delay = (transmitted / (long double) throttling_bandwidth_ul)
					- (long double) (ended.tv_sec - started.tv_sec);
				if (delay > (long double) MAX_THROTTLING_DELAY) {
					(void) gettimeofday(&started, NULL);
					transmitted = (off_t) 0;
                    delay = (long double) MAX_THROTTLING_DELAY;                    
				}
                if (delay > 0.0L) {
					usleep2(delay * 1000000.0L);
				}
			}
#endif														
			filesize += (off_t) r;
# ifdef FTPWHO
			/* Don't update download_total_size, it'd be useless */
            if (shm_data_cur != NULL) {
                shm_data_cur->download_current_size = filesize;
            }
# endif
			
			while (r) {
				ssize_t w;
                
#ifndef WITHOUT_ASCII                
				if (type == 1) {       /* Fuckin ASCII conversion */
                    size_t i = (size_t) 0U;
                    size_t j = (size_t) 0U;
                    size_t k = (size_t) 0U;
                    char cpy[window];

					do {
						if (p[i] == '\r') {
							i++;
							k++;
						} else {
                            cpy[j++] = buf[i++];   /* no overflow, sizeof cpy == sizeof buf */
                        }
					} while (i < (size_t) r);
					r -= k;
                    if (r > 0) {
                        w = write(f, cpy, (size_t) r);
                    } else {
                        w = (ssize_t) 0;
                    }
				} else
#endif				
				{
					w = write(f, p, (size_t) r);
				}

				if (w < 0) {
					close(f);
					xferfd = -1;
					close(c);
					error(-450, MSG_WRITE_FAILED);
                    disp_partial_return:
                    if (guest != 0) {
                        unlinkret = unlink(name);
                    }
					addreply(0, "%s %s", name,
							 unlinkret ? MSG_UPLOAD_PARTIAL : MSG_REMOVED);
                    goto end;
				}
				r -= w;
				p += w;
			}
			r = 1;
		} else if (r < 0) {
			close(f);
			if (xferfd >= 0) {
				xferfd = -1;
				close(c);				
				error(-451, MSG_DATA_READ_FAILED);
                goto disp_partial_return;
			}
            goto end;
		}
	} while (r > 0);
	if (FSTATFS(f, &statfsbuf) == 0) {
		double space;

		space = (double) STATFS_FRSIZE(statfsbuf) * 
			(double) STATFS_BAVAIL(statfsbuf);
		if (space > 524288.0) {
			addreply(0, MSG_SPACE_FREE_M, space / 1048576);
		} else {
			addreply(0, MSG_SPACE_FREE_K, space / 1024);
		}
	}		
	close(f);
	xferfd = -1;
	close(c);
	uploaded += filesize - restartat;
	displayrate(MSG_UPLOADED, filesize - restartat, &started, name, 1);
    end:
	restartat = (off_t) 0;	
}

void domdtm(const char *name)
{
	struct stat st;
	struct tm *t;

	if (!name || !*name) {
		addreply(500, MSG_MISSING_ARG);
	} else if (stat(name, &st)) {
#ifdef DEBUG
		if (debug)
			addreply(0, "arg: %s, wd: %s", name, wd);
#endif
		addreply(550, MSG_STAT_FAILURE2);
	} else if (!S_ISREG(st.st_mode)) {
		addreply(550, MSG_NOT_REGULAR_FILE);
	} else {
		t = gmtime((time_t *) & st.st_mtime);
		if (!t) {
			addreply(550, MSG_GMTIME_FAILURE);
		} else {
			addreply(213, "%04d%02d%02d%02d%02d%02d",
					 t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
					 t->tm_hour, t->tm_min, t->tm_sec);
		}
	}
}

void dosize(const char *name)
{
	struct stat st;

	if (!name || !*name) {
		addreply(500, MSG_MISSING_ARG);
	} else if (stat(name, &st)) {
#ifdef DEBUG
		if (debug)
			addreply(0, "arg: %s, wd: %s", name, wd);
#endif
		addreply(550, MSG_STAT_FAILURE2);
	} else if (!S_ISREG(st.st_mode)) {
		addreply(550, MSG_NOT_REGULAR_FILE);
	} else {
		addreply(213, "%llu", (unsigned long long) st.st_size);
	}
}

void dotype(const char *arg)
{
	replycode = 200;			/* bloody awful hack */

	if (!arg || !*arg) {
		addreply(501, MSG_MISSING_ARG "\n" "A(scii) I(mage) L(ocal)");
	} else if (tolower(*arg) == 'a')
		type = 1;
	else if (tolower(*arg) == 'i')
		type = 2;
	else if (tolower(*arg) == 'l') {
		if (arg[1] == '8') {
			type = 2;
		} else if (isdigit(arg[1])) {
			addreply(504, MSG_TYPE_8BIT_FAILURE);
		} else {
			addreply(0, MSG_MISSING_ARG);
			type = 2;
		}
	} else {
		addreply(504, MSG_TYPE_UNKNOWN ": %s", arg);
	}

	addreply(0, MSG_TYPE_SUCCESS " %s", (type > 1) ? "8-bit binary" : "ASCII");
}

void dostru(const char *arg)
{
	if (!arg || !*arg)
		addreply(500, MSG_MISSING_ARG);
	else if (strcasecmp(arg, "F"))
		addreply(504, MSG_STRU_FAILURE);
	else
		addreply(200, "F OK");
}

void domode(const char *arg)
{
	if (!arg || !*arg)
		addreply(500, MSG_MISSING_ARG);
	else if (strcasecmp(arg, "S"))
		addreply(504, MSG_MODE_FAILURE);
	else
		addreply(200, "S OK");
}

void dornfr(char *name)
{
	struct stat st;
	if (guest) {
		addreply(550, MSG_ANON_CANT_RENAME);
		return;
	}
	if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
		addreply(550, MSG_SANITY_FILE_FAILURE);
		return;
	}
	if ((lstat(name, &st)) == 0) {
		if (renamefrom) {
			addreply(0, MSG_RENAME_ABORT);
			(void) free(renamefrom);
		}
		renamefrom = strdup(name);
		addreply(350, MSG_RENAME_RNFR_SUCCESS);
	} else {
		addreply(550, MSG_FILE_DOESNT_EXIST);
	}
	return;
}

void dornto(char *name)
{
	struct stat st;
	if (guest) {
		addreply(550, MSG_ANON_CANT_RENAME);
		return;
	}
	if (checknamesanity(name, dot_write_forbidden != 0 ? 0 : dot_ok) != 0) {
		addreply(550, MSG_SANITY_FILE_FAILURE);
		return;
	}		
	if ((lstat(name, &st)) == 0)
		addreply(550, MSG_RENAME_ALREADY_THERE);
	else if (!renamefrom)
		addreply(503, MSG_RENAME_NORNFR);
	else if ((rename(renamefrom, name)) < 0)
		addreply(550, MSG_RENAME_FAILURE ": %s", strerror(errno));
	else
		addreply(250, MSG_RENAME_SUCCESS);

	if (renamefrom) {
		(void) free(renamefrom);
		renamefrom = NULL;
	}	
	return;
}

void error(int n, const char *msg)
{
	const char *e = strerror(errno);
	
	logfile(LOG_ERR, "%s: %s", msg, e);
	addreply(n, "%s: %s", msg, e);
	CORK_OFF(1);
}

void fixlimits(void)
{
#ifdef HAVE_SETRLIMIT
	static struct rlimit lim;
	
	lim.rlim_max = lim.rlim_cur = MAX_CPU_TIME;
	setrlimit(RLIMIT_CPU, &lim);
	lim.rlim_max = lim.rlim_cur = MAX_DATA_SIZE;
	setrlimit(RLIMIT_DATA, &lim);
# ifndef DEBUG
	lim.rlim_max = lim.rlim_cur = 0;
	setrlimit(RLIMIT_CORE, &lim);
# endif
#endif
}

#ifdef COOKIE
void fortune(void)
{
	int fd;
	char *buf;
	char *bufpnt;
	char *bufend;
	struct stat st;
	off_t gl;
	char *fortunepnt;
	char fortune[1024];	

	if (fortunes_file == NULL || *fortunes_file == 0) {
		return;
	}
	if ((fd = open(fortunes_file, O_RDONLY)) < 0) {
		logfile(LOG_ERR, MSG_OPEN_FAILURE, fortunes_file);
		return;
	}
	if (fstat(fd, &st) < 0 ||
		(((S_IRUSR | S_IRGRP | S_IROTH) & st.st_mode) !=
		 (S_IRUSR | S_IRGRP | S_IROTH)) ||
		!(S_ISREG(st.st_mode) || S_ISLNK(st.st_mode)) || st.st_size < 2 ||
		(buf = mmap(NULL, (size_t) st.st_size,
				PROT_READ, MAP_FILE | MAP_SHARED, fd, 
				(off_t) 0)) == MAP_FAILED) {
		close(fd);
		logfile(LOG_ERR, MSG_OPEN_FAILURE, fortunes_file);
		return;		
	}
	gl = (off_t) (rand() % (st.st_size - 1U));
	bufpnt = buf + gl;
	bufend = buf + st.st_size;
	while (bufpnt != buf) {
		if (bufpnt[0] == '\n') {
			if (&bufpnt[-1] != buf && bufpnt[-1] == '%') {
				if (&bufpnt[-2] != buf && bufpnt[-2] == '\n') {
					break;
				}
			}
		}
		bufpnt--;
	}
	if (bufpnt != buf) {
		while (bufpnt != bufend && *bufpnt == '\n') {
			bufpnt++;
		}
	}
	fortunepnt = fortune;
	while (*bufpnt != 0 && bufpnt != bufend &&
		   fortunepnt != &fortune[sizeof fortune - 1U]) {
		if (bufpnt[0] == '\n') {
			if (&bufpnt[1] != bufend && bufpnt[1] == '%') {
				if (&bufpnt[2] != bufend && bufpnt[2] == '\n') {
					break;
				}
			}
		}
		*fortunepnt++ = *bufpnt++;
	}
	while (fortunepnt != fortune && *fortunepnt == '\n') {
		fortunepnt--;
	}
	*fortunepnt = 0;
	addreply(220, "<<");
	addreply(220, "%s", fortune);
	addreply(220, ">>");	
	munmap(buf, st.st_size);
	close(fd);
}
#endif

#if !defined(NO_STANDALONE) && !defined(NO_INETD)
static int check_standalone(void)
{
	socklen_t socksize = (socklen_t) sizeof ctrlconn;
	
	if (getsockname(0, (struct sockaddr *) &ctrlconn, &socksize) != 0) {
		return 1;
	}
	return 0;
}
#endif

static void set_signals_client(void)
{
	struct sigaction sa;
		
	sigemptyset(&sa.sa_mask);
	
	sa.sa_flags = SA_RESTART;	
	sa.sa_handler = sigurg;
	(void) sigaction(SIGURG, &sa, NULL);
	sa.sa_handler = SIG_IGN;
	(void) sigaction(SIGCHLD, &sa, NULL);
	(void) sigaction(SIGPIPE, &sa, NULL);
	
	sa.sa_flags = 0;
	sa.sa_handler = sigalarm;	
	(void) sigaction(SIGALRM, &sa, NULL);	
	sa.sa_handler = sigterm_client;
	(void) sigaction(SIGTERM, &sa, NULL);
	(void) sigaction(SIGHUP, &sa, NULL);	
	(void) sigaction(SIGQUIT, &sa, NULL);
	(void) sigaction(SIGINT, &sa, NULL);
	(void) sigaction(SIGKILL, &sa, NULL);
}

static void set_signals(void)
{
#ifndef NO_STANDALONE
	struct sigaction sa;
		
	sigemptyset(&sa.sa_mask);
			
	sa.sa_flags = SA_RESTART;	
	sa.sa_handler = sigchild;
	(void) sigaction(SIGCHLD, &sa, NULL);
	
	sa.sa_flags = 0;
	sa.sa_handler = sigterm;
	(void) sigaction(SIGTERM, &sa, NULL);
	(void) sigaction(SIGHUP, &sa, NULL);	
	(void) sigaction(SIGQUIT, &sa, NULL);
	(void) sigaction(SIGINT, &sa, NULL);
	(void) sigaction(SIGKILL, &sa, NULL);
#endif
}

void doit(void)
{
	socklen_t socksize;
	struct rusage ru;
	time_t ti = time(NULL);
	struct tm *t;
	
	fixlimits();
#ifdef F_SETOWN
	{
		const pid_t p = getpid();
		fcntl(0, F_SETOWN, p);
		fcntl(1, F_SETOWN, p);
	}
#endif	
	set_signals_client();
	umask((mode_t) 0);
	socksize = (socklen_t) sizeof ctrlconn;
	if (getsockname(0, (struct sockaddr *) &ctrlconn, &socksize) != 0) {
		die(421, LOG_ERR, MSG_NO_SUPERSERVER);
	}
	fourinsix(&ctrlconn);
	if (checkvalidaddr(&ctrlconn) == 0) {
		die(425, LOG_ERR, MSG_INVALID_IP);
	}
	if (STORAGE_FAMILY(ctrlconn) == AF_INET6) {
		serverport = ntohs((unsigned short int) STORAGE_PORT6(ctrlconn));
	} else {
		serverport = ntohs((unsigned short int) STORAGE_PORT(ctrlconn));
	}
    if (trustedip != NULL && addrcmp(&ctrlconn, trustedip) != 0) {
       anon_only = 1; 
    }
	socksize = (socklen_t) sizeof peer;
	if (getpeername(0, (struct sockaddr *) &peer, &socksize)) {
		die(421, LOG_ERR, MSG_GETPEERNAME ": %s" , strerror(errno));
	}
	fourinsix(&peer);
	if (checkvalidaddr(&peer) == 0) {
		die(425, LOG_ERR, MSG_INVALID_IP);
	}	
	getnameinfo((struct sockaddr *) &peer, STORAGE_LEN(peer), host,
				sizeof host, NULL, (size_t) 0U, 
				resolve_hostnames != 0 ? 0 : NI_NUMERICHOST);
	iptropize(&peer);
	logfile(LOG_INFO, MSG_NEW_CONNECTION, host);
	
#ifndef NO_BANNER
	addreply(0,
			 "=(<*>)=-.:. (( " MSG_WELCOME_TO " PureFTPd " VERSION
			 " )) .:.-=(<*>)=-");
#else
	{
		char hostname[NI_MAXHOST];
		const char *hostname_ = "";
		
		if (gethostname(hostname, sizeof hostname - 1U) >= 0) {
			/* According to SUSv2, the host name may be not nul-terminated */
			hostname[sizeof hostname - 1U] = 0;
			hostname_ = hostname;
		}
		addreply(220, "%s FTP server ready.", hostname_);		
	}
#endif
	
	if (maxusers > 0U) {
		unsigned int users;

#ifdef NO_STANDALONE
		users = daemons(serverport);
#else
# ifdef NO_INETD
		users = nb_children;
# else
		if (standalone) {
			users = nb_children;
		} else {
			users = daemons(serverport);
		}
# endif
#endif
		if (users > maxusers) {
			addreply(421, MSG_MAX_USERS, maxusers);
			doreply();
			_EXIT(1);
		}
#ifndef NO_BANNER
		addreply(0, MSG_NB_USERS, users, maxusers);
#endif
	}

	/* It's time to add a new entry to the ftpwho list */
#ifdef FTPWHO
	{
		size_t dummy_size;
		
		ftpwho_initwho(1, &dummy_size);
		if (shm_data_cur != NULL) {
			ftpwho_lock();			
			shm_data_cur->pid = getpid();
			shm_data_cur->state = FTPWHO_STATE_IDLE;
			shm_data_cur->addr = peer;
			shm_data_cur->local_addr = ctrlconn;
			shm_data_cur->date = time(NULL);
			shm_data_cur->xfer_date = shm_data_cur->date;
			(shm_data_cur->account)[0] = '?';		
			(shm_data_cur->account)[1] = 0;
			shm_data_cur->download_total_size = (off_t) 0;
			shm_data_cur->download_current_size = (off_t) 0;			
			ftpwho_unlock();
		}
	}
#endif

	/* Back to the client */
#ifdef HAVE_GETLOADAVG                 /* BSD */
    if (getloadavg(&load, 1) < 0) {
        load = -1.0;
    }
#else
    { /* Linux */
        FILE *f;
        
        load = -1.0;    
        if ((f = fopen("/proc/loadavg", "rt")) != NULL) {
            fscanf(f, "%lf", &load);
            fclose(f);
        }
    }
#endif
	wd[0] = '/';
	wd[1] = 0;
	if (chdir(wd)) {
		_EXIT(EXIT_FAILURE);
	}
	{
		int fodder;
#ifdef IPTOS_LOWDELAY
		fodder = IPTOS_LOWDELAY;
		setsockopt(0, SOL_IP, IP_TOS, (char *) &fodder, sizeof fodder);
		setsockopt(1, SOL_IP, IP_TOS, (char *) &fodder, sizeof fodder);	
#endif
#ifdef SO_OOBINLINE
		fodder = 1;
		setsockopt(0, SOL_SOCKET, SO_OOBINLINE, (char *) &fodder, sizeof fodder);
#endif
        nokeepalive(0);
        nokeepalive(1);
	}
	srand((unsigned int) ti);
	t = localtime(&ti);	
#ifdef COOKIE
	fortune();
#endif
#ifndef NO_BANNER
	if (t != NULL)
		addreply(220, MSG_WELCOME_TIME,
				 t->tm_hour, t->tm_min, load > 0.0 ? load : 0.0,
                 (unsigned int) serverport);
#endif
	if (anon_only > 0) {
		addreply(220, MSG_ANONYMOUS_FTP_ONLY);
	} else if (anon_only < 0) {
		addreply(220, MSG_NO_ANONYMOUS_LOGIN);	
	}
#ifndef NO_BANNER
	if (allowfxp == 2) {
		addreply(220, MSG_FXP_SUPPORT);
	}
#endif
#ifdef RATIOS
	if (ratio_upload > 0) {
		if (ratio_for_non_anon != 0) {
			addreply(0, MSG_RATIOS_EVERYONE);
		} else {
			addreply(0, MSG_RATIOS_ANONYMOUS);
		}
		addreply(0, MSG_RATIOS_RULE,
				 ratio_download, ratio_upload);
	}
#endif
#ifndef NO_BANNER
	if (idletime >= 120UL) {
		addreply(220,
				 MSG_INFO_IDLE_M,
				 idletime / 60UL);
	} else {
		addreply(220,
				 MSG_INFO_IDLE_S,
				 (unsigned long) idletime);
	}
#endif

	candownload = ((maxload <= 0.0) || (load < maxload));

#ifdef WITH_LDAP
	if (ftp_ldap_init(ldap_server, ldap_port, 
					  ldap_bind_dn, ldap_bind_pw,
					  ldap_base_dn) != 0) {
		die(421, LOG_ERR, MSG_LDAP_WRONG_PARMS);
	}
	getpwnam_wrapper = ftp_ldap_getpwnam;
#endif	
	parser();
	if (getrusage(RUSAGE_SELF, &ru) == 0) {
		unsigned long s, u;

		u = (ru.ru_utime.tv_usec + ru.ru_stime.tv_usec + 500UL) / 1000UL;
		s = ru.ru_utime.tv_sec + ru.ru_stime.tv_sec;
		if (u > 999UL) {
			u -= 1000UL;
			s++;
		}
		addreply(0, MSG_INFO_CPU_TIME, s, u);
		logfile(LOG_INFO, MSG_INFO_CPU_TIME, s, u);
	} else {
		logfile(LOG_INFO, MSG_INFO_CPU_TIME, 42UL, 42UL);
	}
	doreply();
}

static void check_ipv6_support(void) 	/* check for ipv6 support in kernel */
{
	int p;
	
	if ((p = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP)) != -1) {
		close(p);
		v6ready++;
	}
}

#ifndef NO_STANDALONE
static void updatepidfile(void)
{
	const char * const pf = PID_FILE;
	int fd;
	char buf[42];
	
	if (snprintf(buf, sizeof buf, "%lu\n", (unsigned long) getpid()) < 0) {
		return;
	}
	if (unlink(pf) != 0 && errno != ENOENT) {
		return;
	}
	if ((fd = open(pf, O_CREAT | O_WRONLY | O_TRUNC |
				   O_NOFOLLOW | O_EXCL, 0644)) < 0) {
		return;
	}
	write(fd, buf, strlen(buf));
	close(fd);
}

#ifndef NO_STANDALONE
static void dodaemonize(void)
{
	pid_t child;	
	
	/* Contributed by Jason Lunz - also based on APUI code, see open_max() */		
	if (daemonize) {
		if ((child = fork()) < (pid_t) 0) {
			perror(MSG_STANDALONE_FAILED " - fork");
			logfile(LOG_ERR, MSG_STANDALONE_FAILED " : [fork: %s]", strerror(errno));
			return;
		} else if (child != (pid_t) 0) {
			_EXIT(EXIT_SUCCESS);	   /* parent exits */
		} else {
			int i = open_max();
			
			if (setsid() == (pid_t) -1) {
				perror(MSG_STANDALONE_FAILED " - setsid");   /* continue anyway */
			}
			chdir("/");
			do {
				if (isatty(i)) {
					close(i);
				}
				i--;				
			} while (i != 0);
		}
	}	
}
#endif

static void standalone_server(void)
{
	int clientfd;
	int on;
	struct addrinfo hints, *res;
	struct sockaddr_storage sa;
	int dummy = sizeof sa;
	pid_t child;

	if (isatty(0)) {
		close(0);
	}
# ifndef NO_INETD
	standalone = 1;
# endif
	memset(&hints, 0, sizeof hints);
	hints.ai_flags = AI_PASSIVE;
	if (v6ready) {
		hints.ai_family = AF_INET6;
	} else {
		hints.ai_family = AF_INET;
	}
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	hints.ai_addr = NULL;	
	if ((on = getaddrinfo(standalone_ip, standalone_port, &hints, &res)) != 0) {
		if (v6ready) {
			hints.ai_family = AF_INET;
			if ((on = getaddrinfo(standalone_ip, standalone_port, &hints, &res)) != 0) {
				axe:
				fprintf(stderr, MSG_STANDALONE_FAILED " - " MSG_SYNTAX_ERROR_IP " : %d\n", on);
				logfile(LOG_ERR, MSG_SYNTAX_ERROR_IP " : [%s]", strerror(errno));
				return;				
			}
		} else {
			goto axe;
		}		
	}
	on = 1;	
	if ((listenfd = socket(res->ai_family, SOCK_STREAM, IPPROTO_TCP)) < 0 ||
		setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on) != 0 ||
		bind(listenfd, res->ai_addr, (socklen_t) res->ai_addrlen) != 0 ||
		listen(listenfd, maxusers > 0U ? 3U + maxusers / 8U : DEFAULT_BACKLOG) != 0) {
		freeaddrinfo(res);
		perror(MSG_STANDALONE_FAILED);
		logfile(LOG_ERR, MSG_STANDALONE_FAILED " : [%s]", strerror(errno));
		return;
	}
	freeaddrinfo(res);		
	set_cloexec_flag(listenfd);
	updatepidfile();
	setprogname("pure-ftpd [SERVER]");
	while (stop_server == 0) {
		sigset_t set;		
		
		memset(&sa, 0, sizeof sa);
		if ((clientfd = accept(listenfd, (struct sockaddr *) &sa, &dummy)) < 0) {
			continue;
		}
		if (STORAGE_FAMILY(sa) != AF_INET
			&& STORAGE_FAMILY(sa) != AF_INET6) {
			close(clientfd);
			continue;
		}
		if (maxusers > 0U && nb_children >= maxusers) {
			char line[1024];
			
			if (snprintf(line, sizeof line, "421 " MSG_MAX_USERS "\r\n", (unsigned long) maxusers) > 0) {
				write(clientfd, line, strlen(line));
			}
			close(clientfd);
			continue;
		}
		if (maxip > 0U) {
			fourinsix(&sa);
			if (iptrack_get(&sa) >= maxip) {
				char line[1024];
				char hbuf[NI_MAXHOST];
				static struct sockaddr_storage old_sa;

				if (snprintf(line, sizeof line, "421 " MSG_MAX_USERS_IP "\r\n", (unsigned long) maxip) > 0) {
					write(clientfd, line, strlen(line));
				}
				if (addrcmp(&old_sa, &sa) != 0) {
					old_sa = sa;
					getnameinfo((struct sockaddr *) &sa, STORAGE_LEN(sa), hbuf,
								sizeof hbuf, NULL, (size_t) 0U,
								resolve_hostnames != 0 ? 0 : NI_NUMERICHOST);
					logfile(LOG_WARNING, MSG_MAX_USERS_IP " : [%s]",
							(unsigned long) maxip, hbuf);
				}
				close(clientfd);				
				continue;
			}
		}
		sigemptyset(&set);
		sigaddset(&set, SIGCHLD);
		sigprocmask(SIG_BLOCK, &set, NULL);
		nb_children++;		
		child = fork();
		if (child == (pid_t) 0) {			
			dup2(clientfd, 0);
			dup2(0, 1);
			if (clientfd > 1) {
				close(clientfd);
			}
			if (listenfd > 1) {
				close(listenfd);
			}			
			if (isatty(2)) {
				close(2);
			}
# ifndef SAVE_DESCRIPTORS
			if (no_syslog == 0) {
				closelog();
				openlog("pure-ftpd", LOG_NDELAY, syslog_facility);
			}
# endif
			doit();
			_EXIT(EXIT_SUCCESS);			
		} else if (child < (pid_t) 0) {
			if (nb_children > 0U) {
				nb_children--;
			}
		} else {
			if (maxip > 0U) {
				iptrack_add(&sa, child);
			}
		}
		close(clientfd);
		sigprocmask(SIG_UNBLOCK, &set, NULL);
	}		
}
#endif

int main(int argc, char *argv[])
{
#ifndef NO_GETOPT_LONG
	int option_index = 0;
#endif
	int fodder;
	int bypass_ipv6 = 0;
	struct passwd *pw;

#ifndef HAVE_SETPROCTITLE
	prg_name = argv[0];
#endif

#ifndef SAVE_DECRIPTORS
	openlog("pure-ftpd", LOG_NDELAY, DEFAULT_FACILITY);
#endif

#ifdef USE_CAPABILITIES
	set_initial_caps();
#endif
	set_signals();
	
	loggedin = 0;
	
	while ((fodder =
#ifndef NO_GETOPT_LONG
			getopt_long(argc, argv, GETOPT_OPTIONS, long_options, &option_index)
#else
			getopt(argc, argv, GETOPT_OPTIONS)
#endif
			) != -1) {
		switch (fodder) {
		case 's': {
			if ((pw = getpwnam_wrapper("ftp"))) {
				warez = pw->pw_uid;
			} else {
				logfile(LOG_ERR, MSG_NO_FTP_ACCOUNT);
			}
			break;
		}
		case '4': {
			bypass_ipv6 = 1;
			break;
		}
#ifndef NO_STANDALONE
		case 'S': {
			char *struck;
			
			if ((struck = strchr(optarg, ',')) != NULL) {
				*struck = 0;
				if (*optarg != 0) {
					standalone_ip = strdup(optarg);
				}
				if (struck[1] != 0) {
					standalone_port =  strdup(struck + 1);
				}
			} else {
				standalone_port = strdup(optarg);
			}
			break;
		}		
#endif
		case 'D': {
			force_ls_a = 1;
			break;
		}
#ifdef VUSERS
		case 'v': {
			vusers_file = strdup(optarg);
			break;
		}
#endif
#ifdef THROTTLING
		case 't':
		case 'T': {
			char *struck;
			char *tr_bw_ul = NULL;
			char *tr_bw_dl = NULL;

			if ((struck = strchr(optarg, ':')) != NULL) {
				*struck = 0;
				if (*optarg != 0) {
					tr_bw_ul = optarg;
				}
				if (struck[1] != 0) {
					tr_bw_dl = &struck[1];
				}
			} else {
				tr_bw_ul = tr_bw_dl = optarg;
			}
			if ((tr_bw_ul == NULL || *tr_bw_ul == 0) &&
				(tr_bw_dl == NULL || *tr_bw_dl == 0)) {
				bad_bw:
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_THROTTLING ": %s" , optarg);
			}
			if (tr_bw_dl != NULL) {
				if ((throttling_bandwidth_dl = 
					 strtoul(tr_bw_dl, NULL, 0) * 1024UL) == 0UL) {
					goto bad_bw;
				}
			}
			if (tr_bw_ul != NULL) {
				if ((throttling_bandwidth_ul =
					 strtoul(tr_bw_ul, NULL, 0) * 1024UL) == 0UL) {
					goto bad_bw;
				}
			}
			throttling_delay = 1000000 /
				(throttling_bandwidth_dl | throttling_bandwidth_ul);
			if (fodder == 't') {
				throttling = 1;
			} else {
				throttling = 2;
			}
			break;
		}
#endif
		case 'a': {
			char *nptr, *endptr;

			nptr = optarg;
			endptr = NULL;
			chroot_trustedgid = strtoul(nptr, &endptr, 0);
			if (!nptr || !*nptr || !endptr || *endptr) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_TRUSTED_GID ": %s" , optarg);
			}
			userchroot = 1;
			break;
		} 
		case 'x': {
			dot_write_forbidden = 1;
			break;
		} 
		case 'X': {
			dot_write_forbidden = dot_read_forbidden = 1;
			break;
		} 
		case 'z': {
			dot_read_anon_ok = 1;
			break;
		} 
		case 'A': {
			userchroot = 2;
			break;
		}
		case 'w': {
			allowfxp = 1;
			break;
		} 
		case 'W': {
			allowfxp = 2;
			break;
		} 
		case 'd': {
			if (logging < 2) {
				logging++;
			}
			break;
		} 
		case 'b': {
			broken = 1;
			break;
		} 
		case 'c': {
			char *nptr, *endptr;

			nptr = optarg;
			endptr = NULL;
			maxusers = (unsigned int) strtoul(nptr, &endptr, 0);
			if (!nptr || !*nptr || !endptr || *endptr || !maxusers) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_USER_LIMIT ": %s" , optarg);
			}
			break;
		}
#ifndef NO_STANDALONE
		case 'B': {
			daemonize = 1;
			break;
		} 
		case 'C': {
			char *nptr, *endptr;

			nptr = optarg;
			endptr = NULL;
			maxip = (unsigned int) strtoul(nptr, &endptr, 0);
			if (!nptr || !*nptr || !endptr || *endptr || !maxip) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_USER_LIMIT ": %s" , optarg);
			}
			break;
		}
#endif
		case 'e': {
			anon_only = 1;
			break;
		} 
		case 'E': {
			anon_only = -1;
			break;
		}
#ifdef COOKIE
		case 'F': {
			fortunes_file = strdup(optarg);
			break;
		}
#endif
		case 'f': {
			int n = 0;

			if (strcasecmp(optarg, "none") == 0) {
				no_syslog = 1;
				break;
			}
			while (facilitynames[n].c_name &&
				   strcasecmp(facilitynames[n].c_name, optarg)) {
				n++;
			}
			if (facilitynames[n].c_name) {
				syslog_facility = facilitynames[n].c_val;
			} else {
				logfile(LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_FACILITY ": %s", optarg);
			}
			break;
		}
#ifdef WITH_LDAP
		case 'l': {
			if (generic_parser(optarg, ldap_config_keywords) != 0) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_CONFIG_FILE_LDAP ": %s" , optarg);
			}
			break;
		}
#endif
		case 'm': {
			char *nptr, *endptr;

			nptr = optarg;
			endptr = NULL;
			maxload = strtod(nptr, &endptr);
			if (!nptr || !*nptr || !endptr || *endptr || maxload <= 0.0) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_LOAD_LIMIT ": %s" , optarg);
			}
			break;
		} 
		case 'M': {
			allow_anon_mkdir = 1;
			break;
		}
		case 'N': {
			disallow_passive = 1;
			break;
		}
#if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
		case 'o': {		
			do_upload_script = 1;
			break;
		}
#endif
		case 'p': {
			int ret;

			ret = sscanf(optarg, "%u:%u", &firstport, &lastport);
			if (ret != 2 ||
				firstport < 1024U || lastport > 65535U
				|| lastport < firstport) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_PORTS_RANGE ": %s" , optarg);
			}
			break;
		} 
		case 'L': {
			int ret;

			ret = sscanf(optarg, "%u:%u", &max_ls_files, &max_ls_depth);
			if (ret != 2 ||
				max_ls_files < 1U || max_ls_depth < 1U) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_LS_LIMITS ": %s" , optarg);
			}
			break;
		}
		case 'P': {			
			if (generic_aton(optarg, &force_passive_ip) != 0) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_FORCE_PASSIVE);
			}
			break;
		}
#ifdef RATIOS
		case 'q':
		case 'Q': {
			int ret;

			ret = sscanf(optarg, "%u:%u", &ratio_upload, &ratio_download);
			if (ret != 2 ||
				ratio_upload < 1U || ratio_download < 1U) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_RATIO ": %s" , optarg);
			}
			if (fodder == 'Q') {
				ratio_for_non_anon = 1;
			}
			break;
		}
#endif
        case 'r': {
            autorename = 1;
            break;
        }
		case 'H': {
			resolve_hostnames = 0;
			break;
		} 
		case 'I': {
			char *nptr, *endptr;

			nptr = optarg;
			endptr = NULL;
			idletime = strtoul(nptr, &endptr, 0) * 60UL;
			if (idletime <= 0) {
				idletime = DEFAULT_IDLE;
			}
			break;
		} 
        case 'i': {
            anon_noupload = 1;
            break;
        }
		case 'k': {
			char *nptr, *endptr;

			nptr = optarg;
			endptr = NULL;
			maxdiskusagepct = 1.0 - (strtod(nptr, &endptr) / 100.0);
			if (maxdiskusagepct >= 1.0 || maxdiskusagepct < 0.0) {
				maxdiskusagepct = 0.0;
			}
			break;
		} 
		case 'u': {
			char *nptr, *endptr;
			long tmp;

			nptr = optarg;
			endptr = NULL;
			tmp = strtol(nptr, &endptr, 10);
			if (!nptr || !*nptr || !endptr || *endptr ||
				tmp > 65535 || tmp < 0) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_UID_LIMIT ": %s" , optarg);
			}
			useruid = (uid_t) tmp;
			break;
		} 
		case 'U': {
			char *struck;
			char *tr_umask = NULL;
			char *tr_umask_d = NULL;
			
			if ((struck = strchr(optarg, ':')) != NULL) {
				*struck = 0;
				if (*optarg != 0) {
					tr_umask = optarg;
				}
				if (struck[1] != 0) {
					tr_umask_d = &struck[1];
				}
			} else {
				tr_umask = tr_umask_d = optarg;
			}
			if ((tr_umask == NULL || *tr_umask == 0) &&
				(tr_umask_d == NULL || *tr_umask_d == 0)) {
				bad_umask:
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_UMASK ": %s" , optarg);				
			}
			if (tr_umask != NULL) {
				if ((u_mask = 
					 strtoul(tr_umask, NULL, 8)) > 0777) {
					goto bad_umask;
				}
			}
			if (tr_umask_d != NULL) {
				if ((u_mask_d =
					 strtoul(tr_umask_d, NULL, 8)) > 0777) {
					goto bad_umask;
				}
			}
			break;
		}
#ifdef WITH_VIRTUAL_HOSTS
        case 'V': {
            if ((trustedip = malloc(sizeof *trustedip)) == NULL) {
                die_mem();
            }
			if (generic_aton(optarg, trustedip) != 0) {
				die(421, LOG_ERR, MSG_CONF_ERR ": " MSG_ILLEGAL_TRUSTED_IP);
			}
            break;
        }            
#endif
#ifndef NO_GETOPT_LONG
		case 'h': {
			const struct option *options = long_options;
			
			do {
				printf("-%c\t--%s\t%s\n", options->val, options->name,
					   options->has_arg ? "<opt>" : "");
				options++;
			} while (options->name != NULL);
			exit(EXIT_SUCCESS);
		}
#endif
		default: 
			logfile(LOG_WARNING, MSG_ILLEGAL_OPTION);		
		}
	}
#ifndef SAVE_DECRIPTORS
	if (no_syslog == 0 && syslog_facility != DEFAULT_FACILITY) {
		closelog();
		openlog("pure-ftpd", LOG_NDELAY, syslog_facility);
	}
#endif
#ifndef NO_STANDALONE
	dodaemonize();
#endif
	umask((mode_t) 0);	
	clearargs(argv);
	idletime_noop = (double) idletime * 2.0;
#ifdef WITH_LDAP
	if (ldap_server == NULL) {
		if ((ldap_server = strdup(LDAP_DEFAULT_SERVER)) == NULL) {
			die_mem();
		}
	}
	if (ldap_port_s == NULL) {
		ldap_port = LDAP_DEFAULT_PORT;
	} else {
		ldap_port = atoi(ldap_port_s);
		if (ldap_port <= 0 || ldap_port > 65535) {
			ldap_port = LDAP_DEFAULT_PORT;
		}
	}
	free(ldap_port_s);
	ldap_port_s = NULL;	
	if (ldap_base_dn == NULL) {
		die(421, LOG_ERR, MSG_LDAP_MISSING_BASE);
	}
	if (ldap_bind_pw == NULL || ldap_bind_dn == NULL) {
		ldap_bind_pw = NULL;
		ldap_bind_dn = NULL;
	}
#endif
	if (firstport) {
		unsigned int portmax;
		
		portmax = (lastport - firstport + 1) / 2;
		if (!maxusers || maxusers > portmax) {
			maxusers = portmax;	/* ... so we don't run out of ports */
		}
	}	
	if (bypass_ipv6 == 0) {
		check_ipv6_support();
	}
#if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
	if (do_upload_script != 0) {
		upload_pipe_open();
	}
#endif	
#if !defined(NO_STANDALONE) && !defined(NO_INETD)
	if (check_standalone() != 0) {
		standalone_server();
	} else {
		doit();
	}
#elif !defined(NO_STANDALONE) && defined(NO_INETD)
	standalone_server();
#elif defined(NO_STANDALONE) && !defined(NO_INETD)
	doit();
#else
# error Configuration error
#endif

#if defined(WITH_UPLOAD_SCRIPT) && !defined(NON_ROOT_FTP)
	upload_pipe_close();
#endif
#ifdef WITH_LDAP
	ftp_ldap_exit();
#endif
#ifdef WITH_LDAP
	if (ldap_server != NULL) {
		free(ldap_server);
	}
	if (ldap_port_s != NULL) {
		free(ldap_port_s);
	}
	if (ldap_base_dn != NULL) {
		free(ldap_base_dn);
	}
	if (ldap_bind_dn != NULL) {
		free(ldap_bind_dn);
	}
	if (ldap_bind_pw != NULL) {
		free(ldap_bind_pw);
	}
#endif
    if (trustedip != NULL) {
        free(trustedip);
    }
	unlink(PID_FILE);
	closelog();
#ifdef FTPWHO
	_EXIT(EXIT_SUCCESS);
#endif

	return 0;
}
