
/* LDAP posixAccount handler for Pure-FTPd */
/* (C)opyleft 2001 by Frank DENIS <j@4u.net> */

#include <config.h>
#ifdef WITH_LDAP
# include "ftpd.h"
# include "log_ldap_p.h"
# include "log_ldap.h"
# ifdef WITH_DMALLOC
#  include <dmalloc.h>
# endif

static const char *host;
static int port;
static const char *root;
static const char *pwd;
static const char *base;

int ftp_ldap_init(const char *host_, const int port_,
				  const char *root_, const char *pwd_,
				  const char *base_)
{
	if (host_ != NULL && *host_ != 0) {
		if (host == NULL || strcasecmp(host_, host) != 0) {
			if ((host = strdup(host_)) == NULL) {
				return -1;
			}
		}
	} else {
		host = strdup("localhost");
	}
	if (port_ >= 0 && port_ <= 65535) {
		port = port_;
	} else {
		port = 389;
	}
	if (pwd_ != NULL && *pwd_ != 0) {
		if (pwd == NULL || strcasecmp(pwd_, pwd) != 0) {
			if ((pwd = strdup(pwd_)) == NULL) {
				return -1;
			}
		}
	} else {
		pwd = strdup("");
	}
	if (root_ != NULL && *root_ != 0) {
		if (root == NULL || strcasecmp(root_, root) != 0) {
			if ((root = strdup(root_)) == NULL) {
				return -1;
			}
		}
	} else {
		root = NULL;
		pwd = NULL;
	}
	if (base_ != NULL && *base_ != 0) {
		if (base == NULL || strcasecmp(base_, base) != 0) {
			if ((base = strdup(base_)) == NULL) {
				return -1;
			}
		}
	} else {
		base = strdup("");
	}	
	return 0;
}

int ftp_ldap_exit(void)
{
	if (host != NULL) {
		free((void *) host);
		host = NULL;
	}
	port = -1;
	if (root != NULL) {
		free((void *) root);
		root = NULL;
	}
	if (pwd != NULL) {
		free((void *) pwd);
		pwd = NULL;
	}
	if (base != NULL) {
		free((void *) base);
		base = NULL;
	}
	return 0;
}

static LDAP *ftp_ldap_connect(void)
{
	LDAP *ld;

	if (host == NULL || port < 0) {
		return NULL;
	}
	if ((ld = ldap_init(host, port)) == NULL) {
		return NULL;
	}
	if (ldap_bind_s(ld, root, pwd, LDAP_AUTH_SIMPLE) != LDAP_SUCCESS) {
		return NULL;
	}
	
	return ld;
}

static LDAPMessage *ftp_ldap_uid_search(LDAP * const ld, 
										const char *uid,
										char *attrs[])
{
	char *filter;
	size_t uid_size;
	size_t filter_size;
	int rc;	
	LDAPMessage *res;
	
	if (uid == NULL || *uid == 0) {
		return NULL;
	}
	uid_size = strlen(uid);
	if (uid_size > MAX_LDAP_UID_LENGTH) {
		return NULL;
	}
	filter_size = uid_size + 
		(sizeof  "(&(objectClass=posixAccount)(uid=" - 1U) + 
		(sizeof "))" -1U) + 1U;
	if ((filter = ALLOCA(filter_size)) == NULL) {
		return NULL;
	}
	if (snprintf(filter, filter_size, 
				 "(&(objectClass=posixAccount)(uid=%s))", uid) <= 0) {
		ALLOCA_FREE(filter);
		return NULL;
	}
	rc = ldap_search_s(ld, base, LDAP_SCOPE_SUBTREE, filter, attrs, 0, &res);
	ALLOCA_FREE(filter_size);
	if (rc != LDAP_SUCCESS) {
		return NULL;
	}
	return res;
}

static char *ftp_ldap_getvalue(LDAP * const ld,
							   LDAPMessage * const res,
							   const char * const attribute)
{
	char **vals;
	char *ret;
	
	if ((vals = ldap_get_values(ld, res, attribute)) == NULL ||
		vals[0] == NULL) {
		return NULL;
	}	
	ret = strdup(vals[0]);
	ldap_value_free(vals);
	
	return ret;
}

static void ftp_ldap_getpwnam_freefields(struct passwd * const p) 
{
	if (p->pw_passwd != NULL) {
		free(p->pw_passwd);
		p->pw_passwd = NULL;
	}
	if (p->pw_dir != NULL) {
		free(p->pw_dir);
		p->pw_dir = NULL;
	}
	if (p->pw_shell != NULL) {
		free(p->pw_shell);
		p->pw_shell = NULL;
	}
}

static int ftp_ldap_validate_name(const char *name)
{
	if (name == NULL || *name == 0) {
		return -1;
	}
	do {
		if ((*name >= 'a' && *name <= 'z') ||
			(*name >= 'A' && *name <= 'Z') ||
			(*name >= '0' && *name <= '9') ||
			*name == ' ' || *name == '-' ||
			*name == '_' || *name == '\'' || *name == '.') {
			/* God bless the Perl 'unless' keyword */
		} else {
			return -1;
		}			
		name++;
	} while (*name != 0);
	
	return 0;
}

struct passwd *ftp_ldap_getpwnam(const char *name)
{
	static struct passwd pwret = {
		NULL, NULL, (uid_t) -1, (gid_t) -1,
			NULL, NULL, NULL
	};
	LDAP *ld;
	LDAPMessage *res;
	char *attrs[] = { 
		"uidNumber", "gidNumber", "homeDirectory",
			"userPassword", "loginShell", NULL
	};
	const char *pw_uid_s = NULL;
	const char *pw_gid_s = NULL;
	const char *pw_passwd_ldap = NULL;

	if (ftp_ldap_validate_name(name) != 0) {
		return NULL;
	}
	if ((ld = ftp_ldap_connect()) == NULL) {
		return NULL;
	}
	if ((res = ftp_ldap_uid_search(ld, name, attrs)) == NULL) {
		goto error;
	}
	ftp_ldap_getpwnam_freefields(&pwret);	
	pwret.pw_name = (char *) name;
	if ((pw_passwd_ldap = 
		 ftp_ldap_getvalue(ld, res, "userPassword")) == NULL) {
		goto error;
	}
	if (strncasecmp(pw_passwd_ldap, PASSWD_LDAP_CRYPT_PREFIX,
					sizeof PASSWD_LDAP_CRYPT_PREFIX - 1U) == 0) {
		pwret.pw_passwd = strdup(pw_passwd_ldap + 
								 sizeof PASSWD_LDAP_CRYPT_PREFIX - 1U);		
	} else if (strncasecmp(pw_passwd_ldap, PASSWD_LDAP_MD5_PREFIX,
						   sizeof PASSWD_LDAP_MD5_PREFIX - 1U) == 0) {
		pwret.pw_passwd = malloc(sizeof "$1$" - 1U +
								 strlen(pw_passwd_ldap) -
								 (sizeof PASSWD_LDAP_MD5_PREFIX - 1U)
								 + 1U);
		pwret.pw_passwd[0] = '$';
		pwret.pw_passwd[1] = '1';
		pwret.pw_passwd[2] = '$';
		strcpy(&pwret.pw_passwd[3], /* Ok, no overflow */
			   pw_passwd_ldap + sizeof PASSWD_LDAP_MD5_PREFIX - 1U); 
	} else {
		pwret.pw_passwd = strdup(pw_passwd_ldap);
	}
	free((void *) pw_passwd_ldap);
	if (pwret.pw_passwd == NULL) {
		goto error;
	}	
	pw_passwd_ldap = NULL;
	if ((pw_uid_s =
		 ftp_ldap_getvalue(ld, res, "uidNumber")) == NULL) {
		goto error;
	}
	pwret.pw_uid = (uid_t) strtoul(pw_uid_s, NULL, 10);
	free((void *) pw_uid_s);
	pw_uid_s = NULL;
	if ((pw_gid_s =
		 ftp_ldap_getvalue(ld, res, "gidNumber")) == NULL) {
		goto error;
	}
	pwret.pw_gid = (gid_t) strtoul(pw_gid_s, NULL, 10);
	free((void *) pw_gid_s);
	pw_gid_s = NULL;
	if ((pwret.pw_dir = 
		 ftp_ldap_getvalue(ld, res, "homeDirectory")) == NULL) {
		goto error;
	}	
	if ((pwret.pw_shell = 
		 ftp_ldap_getvalue(ld, res, "loginShell")) == NULL) {
		pwret.pw_shell = strdup(DEFAULT_SHELL);
	}

	ldap_msgfree(res);
	ldap_unbind(ld);	
	
	return &pwret;
	
	error:	
	if (res != NULL) {
		ldap_msgfree(res);
	}
	ldap_unbind(ld);
	ftp_ldap_getpwnam_freefields(&pwret);
	if (pw_uid_s != NULL) {
		free((void *) pw_uid_s);
	}
	if (pw_gid_s != NULL) {
		free((void *) pw_gid_s);
	}	
	if (pw_passwd_ldap != NULL) {
		free((void *) pw_passwd_ldap);
	}	

	return NULL;
}
#else
extern signed char v6ready;
#endif
